# !/usr/bin/env python
# -*- coding: utf-8 -*-

import click
import sys

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.stats import PairCounter
from ..lib.parse_pysam import AlignmentFilePairtoolized
from ..lib.parse import streaming_classify

UTIL_NAME = "pairtools_parse2"


@cli.command()
@click.argument("sam_path", type=str, required=False)
# Parsing options:
@click.option(
    "-c",
    "--chroms-path",
    type=str,
    required=True,
    help="Chromosome order used to flip interchromosomal mates: "
    "path to a chromosomes file (e.g. UCSC chrom.sizes or similar) whose "
    "first column lists scaffold names. Any scaffolds not listed will be "
    "ordered lexicographically following the names provided.",
)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file with pairs. "
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4-compressed."
    "By default, the output is printed into stdout. ",
)
@click.option(
    "--report-position",
    type=click.Choice(["junction", "read", "walk", "outer"]),
    default="outer",
    help="""Reported position of alignments in pairs of complex walks (pos columns). 
    Each alignment in .bam/.sam Hi-C-like data has two ends, and you can report one or another depending of the position of alignment on a read or in a pair.  
    
    "junction" - inner ends of sequential alignments in each pair, aka ligation junctions (complex walks default),
    "read" - 5'-end of alignments relative to R1 or R2 read coordinate system (as in traditional Hi-C),
    "walk" - 5'-end of alignments relative to the whole walk coordinate system,
    "outer" - outer ends of sequential alignments in each pair. """,
)
@click.option(
    "--report-orientation",
    type=click.Choice(["pair", "read", "walk", "junction"]),
    default="pair",
    help="""Reported orientataion of pairs in complex walk (strand columns).
    Each alignment in .bam/.sam Hi-C-like data has orientation, and you can report it relative to the read, pair or whole walk coordinate system.
    
    "pair" - orientation as if each pair in complex walk was sequenced independently from the outer ends or molecule (as in traditional Hi-C, also complex walks default),
    "read" - orientation defined by the read (R1 or R2 read coordinate system),
    "walk" - orientation defined by the walk coordinate system,
    "junction" - reversed "pair" orientation, as if pair was sequenced in both directions starting from the junction""",
)
@click.option(
    "--assembly",
    type=str,
    help="Name of genome assembly (e.g. hg19, mm10) to store in the pairs header.",
)
@click.option(
    "--min-mapq",
    type=int,
    default=1,
    show_default=True,
    help="The minimal MAPQ score to consider a read as uniquely mapped.",
)
@click.option(
    "--max-inter-align-gap",
    type=int,
    default=20,
    show_default=True,
    help="Read segments that are not covered by any alignment and"
    ' longer than the specified value are treated as "null" alignments.'
    " These null alignments convert otherwise linear alignments into walks,"
    " and affect how they get reported as a Hi-C pair.",
)
@click.option(
    "--max-insert-size",
    type=int,
    default=500,
    show_default=True,
    help="When searching for overlapping ends of left and right read (R1 and R2), this sets the minimal distance "
    "when two alignments on the same strand and chromosome are considered part of the same fragment (and thus reported as the same alignment "
    "and not a pair). For traditional Hi-C with long restriction fragments and shorter molecules after ligation+sonication, this "
    "can be the expected molecule size. For complex walks with short restriction fragments, this can be the expected restriction fragment "
    "size. Note that unsequenced insert is *terra incognita* and might contain unsequenced DNA (including ligations) in it. "
    "This parameter is ignored in --single-end mode. ",
)
@click.option(
    "--dedup-max-mismatch",
    type=int,
    default=3,
    show_default=True,
    help="Allowed mismatch between intramolecular alignments to detect readthrough duplicates. "
    "Pairs with both sides mapped within this distance (bp) from each "
    "other are considered duplicates. ",
)
@click.option(
    "--single-end",
    is_flag=True,
    help="If specified, the input is single-end. "
    "Never use this for paired-end data, because R1 read will be omitted. "
    "If single-end data is provided, but parameter is unset, the pairs will be "
    "generated, but may contain artificial UN pairs. ",
)
@click.option(
    "--expand/--no-expand",
    is_flag=True,
    help="If specified, perform combinatorial expansion on the pairs. "
    "Combinatorial expansion is a way to increase the number of contacts in you data, assuming that all DNA fragments in the same molecule (read) are in contact. "
    "Expanded pairs have modified pair type, 'E{separation}_{pair type}'",
)
@click.option(
    "--max-expansion-depth",
    type=int,
    default=None,
    show_default=True,
    help="Works in combination with --expand. "
    "Maximum number of segments separating pair. By default, expanding all possible pairs."
    "Setting the number will limit the expansion depth and enforce contacts from the same "
    "side of the read. ",
)
@click.option(
    "--add-pair-index",
    is_flag=True,
    help="If specified, parse2 will report pair index in the walk as additional columns (R1, R2, R1&R2 or R1-R2). "
    "See documentation: https://pairtools.readthedocs.io/en/latest/parsing.html#rescuing-complex-walks "
    "For combinatorial expanded pairs, two numbers will be reported: "
    "original pair index of the left and right segments. ",
)
@click.option(
    "--flip/--no-flip",
    is_flag=True,
    default=False,
    help="If specified, flip pairs in genomic order and instead preserve "
    "the order in which they were sequenced. Note that no flip is recommended for analysis of walks because it will "
    "override the order of alignments in pairs. Flip is required for appropriate deduplication of sorted pairs. "
    "Flip is not required for cooler cload, which runs flipping internally. ",
)
@click.option(
    "--add-columns",
    type=click.STRING,
    default="",
    help="Report extra columns describing alignments "
    "Possible values (can take multiple values as a comma-separated "
    "list): a SAM tag (any pair of uppercase letters) or {}.".format(
        ", ".join(pairsam_format.EXTRA_COLUMNS)
    ),
)
@click.option(
    "--drop-readid/--keep-readid",
    is_flag=True,
    default=False,
    help="If specified, do not add read ids to the output. By default, keep read ids. Useful for long walks analysis. ",
)
@click.option(
    "--readid-transform",
    type=str,
    default=None,
    help="A Python expression to modify read IDs. Useful when read IDs differ "
    "between the two reads of a pair. Must be a valid Python expression that "
    "uses variables called readID and/or i (the 0-based index of the read pair "
    "in the bam file) and returns a new value, e.g. \"readID[:-2]+'_'+str(i)\". "
    "Make sure that transformed readIDs remain unique!",
    show_default=True,
)
@click.option(
    "--drop-seq/--keep-seq",
    is_flag=True,
    default=False,
    help="Remove sequences and PHREDs from the sam fields by default. Kept otherwise. ",
)
@click.option(
    "--drop-sam/--keep-sam",
    is_flag=True,
    default=False,
    help="Do not add sams to the output by default. Kept otherwise. ",
)
@click.option(
    "--output-parsed-alignments",
    type=str,
    default="",
    help="output file with all parsed alignments (one alignment per line)."
    " Useful for debugging and analysis of walks."
    " If file exists, it will be open in the append mode."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4-compressed."
    " By default, not used.",
)
@click.option(
    "--output-stats",
    type=str,
    default="",
    help="output file for various statistics of pairs file. "
    " By default, statistics is not generated.",
)
@common_io_options
def parse2(
    sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
):
    """Extracts pairs from .sam/.bam data with complex walks, make .pairs.
    SAM_PATH : an input .sam/.bam file with paired-end or single-end sequence alignments of
    Hi-C (or Hi-C-like) molecules. If the path ends with .bam, the input is decompressed from
    bam with samtools. By default, the input is read from stdin.
    """
    parse2_py(
        sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
    )


def parse2_py(
    sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
):
    ### Set up input stream
    if sam_path:  # open input sam file with pysam
        input_sam = AlignmentFilePairtoolized(
            sam_path, "r", threads=kwargs.get("nproc_in")
        )
    else:  # read from stdin
        input_sam = AlignmentFilePairtoolized("-", "r", threads=kwargs.get("nproc_in"))

    ### Set up output streams
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )
    out_alignments_stream = (
        fileio.auto_open(
            output_parsed_alignments,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output_parsed_alignments
        else None
    )
    out_stats_stream = (
        fileio.auto_open(
            output_stats,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output_stats
        else None
    )

    if out_alignments_stream:
        out_alignments_stream.write(
            "readID\tside\tchrom\tpos\tstrand\tmapq\tcigar\tdist_5_lo\tdist_5_hi\tmatched_bp\n"
        )

    # generate empty PairCounter if stats output is requested:
    out_stat = PairCounter() if output_stats else None

    ### Set up output parameters
    add_columns = kwargs.get("add_columns", [])
    add_columns = [col for col in add_columns.split(",") if col]
    for col in add_columns:
        if not (
            (col in pairsam_format.EXTRA_COLUMNS) or (len(col) == 2 and col.isupper())
        ):
            raise Exception("{} is not a valid extra column".format(col))

    columns = pairsam_format.COLUMNS + (
        [c + side for c in add_columns for side in ["1", "2"]]
    )

    if kwargs.get("drop_sam", True):
        columns.pop(columns.index("sam1"))
        columns.pop(columns.index("sam2"))

    if not kwargs.get("add_pair_index", False):
        columns.pop(columns.index("walk_pair_index"))
        columns.pop(columns.index("walk_pair_type"))

    ### Parse header
    samheader = input_sam.header

    if not samheader:
        raise ValueError(
            "The input sam is missing a header! If reading a bam file, please use `samtools view -h` to include the header."
        )

    ### Parse chromosome files present in the input
    sam_chromsizes = headerops.get_chromsizes_from_pysam_header(samheader)
    chromosomes = headerops.get_chrom_order(chroms_path, list(sam_chromsizes.keys()))

    ### Write new header to the pairsam file
    header = headerops.make_standard_pairsheader(
        assembly=kwargs.get("assembly", ""),
        chromsizes=[(chrom, sam_chromsizes[chrom]) for chrom in chromosomes],
        columns=columns,
        shape="whole matrix" if not kwargs["flip"] else "upper triangle",
    )

    header = headerops.insert_samheader_pysam(header, samheader)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))

    ### Parse input and write to the outputs
    streaming_classify(
        input_sam,
        outstream,
        chromosomes,
        out_alignments_stream,
        out_stat,
        parse2=True,
        **kwargs
    )

    # save statistics to a file if it was requested:
    if out_stat:
        out_stat.save(out_stats_stream)

    if outstream != sys.stdout:
        outstream.close()
    if out_alignments_stream:
        out_alignments_stream.close()
    if out_stats_stream:
        out_stats_stream.close()


if __name__ == "__main__":
    parse2()



PAIRSAM_FORMAT_VERSION = "1.0.0"

PAIRSAM_SEP = "\t"
PAIRSAM_SEP_ESCAPE = r"\t"
SAM_SEP = "\031"
SAM_SEP_ESCAPE = r"\031"
INTER_SAM_SEP = "\031NEXT_SAM\031"

COL_READID = 0
COL_C1 = 1
COL_P1 = 2
COL_C2 = 3
COL_P2 = 4
COL_S1 = 5
COL_S2 = 6
COL_PTYPE = 7
COL_SAM1 = 8
COL_SAM2 = 9

COLUMNS = [
    "readID",
    "chrom1",
    "pos1",
    "chrom2",
    "pos2",
    "strand1",
    "strand2",
    "pair_type",
    "sam1",
    "sam2",
    "walk_pair_index",
    "walk_pair_type",
]

# Required columns for formats:
COLUMNS_PAIRSAM = [
    "readID",
    "chrom1",
    "pos1",
    "chrom2",
    "pos2",
    "strand1",
    "strand2",
    "pair_type",
    "sam1",
    "sam2",
]

COLUMNS_PAIRS = [
    "readID",
    "chrom1",
    "pos1",
    "chrom2",
    "pos2",
    "strand1",
    "strand2",
    "pair_type",
]

UNMAPPED_CHROM = "!"
UNMAPPED_POS = 0
UNMAPPED_STRAND = "-"

UNANNOTATED_RFRAG = -1

EXTRA_COLUMNS = [
    "mapq",
    "pos5",
    "pos3",
    "cigar",
    "read_len",
    "matched_bp",
    "algn_ref_span",
    "algn_read_span",
    "dist_to_5",
    "dist_to_3",
    "seq",
    "mismatches",  # Format: "{ref_letter}:{mut_letter}:{phred}:{ref_position}:{read_position}"
]



import sys
import click

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options
import warnings

UTIL_NAME = "pairtools_flip"


@cli.command()
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-c",
    "--chroms-path",
    type=str,
    required=True,
    help="Chromosome order used to flip interchromosomal mates: "
    "path to a chromosomes file (e.g. UCSC chrom.sizes or similar) whose "
    "first column lists scaffold names. Any scaffolds not listed will be "
    "ordered lexicographically following the names provided.",
)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@common_io_options
def flip(pairs_path, chroms_path, output, **kwargs):
    """Flip pairs to get an upper-triangular matrix.

    Change the order of side1 and side2 in pairs, such that
    (order(chrom1) < order(chrom2)
     or (order(chrom1) == order(chrom2)) and (pos1 <=pos2))
    Equivalent to reflecting the lower triangle of a Hi-C matrix onto
    its upper triangle, resulting in an upper triangular matrix.
    The order of chromosomes must be provided via a .chromsizes file.

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.

    """
    flip_py(pairs_path, chroms_path, output, **kwargs)


def flip_py(pairs_path, chroms_path, output, **kwargs):

    instream = (
        fileio.auto_open(
            pairs_path,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        if pairs_path
        else sys.stdin
    )
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )

    chromosomes = headerops.get_chrom_order(chroms_path)
    chrom_enum = dict(
        zip(
            [pairsam_format.UNMAPPED_CHROM] + list(chromosomes),
            range(len(chromosomes) + 1),
        )
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))

    column_names = headerops.extract_column_names(header)
    if len(column_names) == 0:
        column_names = pairsam_format.COLUMNS

    chrom1_col = column_names.index("chrom1")
    chrom2_col = column_names.index("chrom2")
    pos1_col = column_names.index("pos1")
    pos2_col = column_names.index("pos2")
    pair_type_col = (
        column_names.index("pair_type") if "pair_type" in column_names else -1
    )

    col_pairs_to_flip = [
        (column_names.index(col), column_names.index(col[:-1] + "2"))
        for col in column_names
        if col.endswith("1") and (col[:-1] + "2") in column_names
    ]

    for line in body_stream:
        cols = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)

        is_annotated1 = cols[chrom1_col] in chrom_enum.keys()
        is_annotated2 = cols[chrom2_col] in chrom_enum.keys()
        if not is_annotated1 or not is_annotated2:
            warnings.warn(f"Unannotated chromosomes in the pairs file!")
        # Flip so that annotated chromosome stands first:
        if is_annotated1 and not is_annotated2:
            has_correct_order = True
        elif is_annotated2 and not is_annotated1:
            has_correct_order = False
        elif not is_annotated1 and not is_annotated2:
            has_correct_order = cols[chrom1_col] < cols[chrom2_col]
        else:  # both are annotated:
            has_correct_order = (chrom_enum[cols[chrom1_col]], int(cols[pos1_col])) <= (
                chrom_enum[cols[chrom2_col]],
                int(cols[pos2_col]),
            )

        # flipping:
        if not has_correct_order:
            for col1, col2 in col_pairs_to_flip:
                if (col1 < len(cols)) and (col2 < len(cols)):
                    cols[col1], cols[col2] = cols[col2], cols[col1]
            if pair_type_col != -1 and pair_type_col < len(cols):
                cols[pair_type_col] = cols[pair_type_col][1] + cols[pair_type_col][0]

        outstream.write(pairsam_format.PAIRSAM_SEP.join(cols))
        outstream.write("\n")

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    flip()



import numpy as np
import warnings

from .dedup import mark_split_pair_as_dup
from . import pairsam_format


def fetchadd(key, mydict):
    key = key.strip()
    if key not in mydict:
        mydict[key] = len(mydict)
    return mydict[key]


def ar(mylist, val):
    return np.array(mylist, dtype={8: np.int8, 16: np.int16, 32: np.int32}[val])


def _filterbycov(c1_in, p1_in, c2_in, p2_in, max_dist, method):
    """
    This is a slow version of the filtering code used for testing purposes only
    Use cythonized version in the future!!
    """

    c1 = np.asarray(c1_in, dtype=int)
    p1 = np.asarray(p1_in, dtype=int)
    c2 = np.asarray(c2_in, dtype=int)
    p2 = np.asarray(p2_in, dtype=int)

    M = np.r_[
        np.c_[c1, p1], np.c_[c2, p2]
    ]  # M is a table of (chrom, pos) with 2*N rows

    assert c1.shape[0] == c2.shape[0]
    N = 2 * c1.shape[0]

    ind_sorted = np.lexsort((M[:, 1], M[:, 0]))  # sort by chromosomes, then positions
    # M[ind_sorted]
    # ind_sorted
    # M, M[ind_sorted]

    if method == "sum":
        proximity_count = np.zeros(
            N
        )  # keeps track of how many molecules each framgent end is close to
    elif method == "max":
        proximity_count = np.zeros(N)
    else:
        raise ValueError("Unknown method: {}".format(method))

    low = 0
    high = 1
    while True:

        # boundary case finish
        if low == N:
            break

        # boundary case  - CHECK
        if high == N:
            low += 1
            high = low + 1
            continue

        # check if "high" is proximal enough to "low"

        # first, if chromosomes not equal, we have gone too far, and the positions are not proximal
        if M[ind_sorted[low], 0] != M[ind_sorted[high], 0]:
            low += 1
            high = low + 1  # restart high
            continue

        # next, if positions are not proximal, increase low, and continue
        elif np.abs(M[ind_sorted[high], 1] - M[ind_sorted[low], 1]) > max_dist:
            low += 1
            high = low + 1  # restart high
            continue

        # if on the same chromosome, and the distance is "proximal enough", add to count of both "low" and "high" positions
        else:
            proximity_count[low] += 1
            proximity_count[high] += 1

        high += 1

    # unsort proximity count
    # proximity_count = proximity_count[ind_sorted]
    proximity_count[ind_sorted] = np.copy(proximity_count)
    # print(M)
    # print(proximity_count)

    # if method is sum of pairs
    if method == "sum":
        pcounts = proximity_count[0 : N // 2] + proximity_count[N // 2 :] + 1
    elif method == "max":
        pcounts = np.maximum(
            proximity_count[0 : N // 2] + 1, proximity_count[N // 2 :] + 1
        )
    else:
        raise ValueError("Unknown method: {}".format(method))

    return pcounts


def streaming_filterbycov(
    method,
    max_dist,
    max_cov,
    sep,
    c1ind,
    c2ind,
    p1ind,
    p2ind,
    s1ind,
    s2ind,
    unmapped_chrom,
    instream,
    outstream,
    outstream_high,
    outstream_unmapped,
    out_stat,
    mark_multi,
):

    # doing everything in memory
    maxind = max(c1ind, c2ind, p1ind, p2ind, s1ind, s2ind)

    # if we do stats in the dedup, we need PAIR_TYPE
    # i do not see way around this:
    if out_stat:
        ptind = pairsam_format.COL_PTYPE
        maxind = max(maxind, ptind)

    c1 = []
    c2 = []
    p1 = []
    p2 = []
    s1 = []
    s2 = []
    line_buffer = []
    cols_buffer = []
    chromDict = {}
    strandDict = {}
    n_unmapped = 0
    n_high = 0
    n_low = 0

    instream = iter(instream)
    while True:
        rawline = next(instream, None)
        stripline = rawline.strip() if rawline else None

        # take care of empty lines not at the end of the file separately
        if rawline and (not stripline):
            warnings.warn("Empty line detected not at the end of the file")
            continue

        if stripline:
            cols = stripline.split(sep)
            if len(cols) <= maxind:
                raise ValueError(
                    "Error parsing line {}: ".format(stripline)
                    + " expected {} columns, got {}".format(maxind, len(cols))
                )

            if (cols[c1ind] == unmapped_chrom) or (cols[c2ind] == unmapped_chrom):

                if outstream_unmapped:
                    outstream_unmapped.write(stripline)
                    # don't forget terminal newline
                    outstream_unmapped.write("\n")

                # add a pair to PairCounter if stats output is requested:
                if out_stat:
                    out_stat.add_pair(
                        cols[c1ind],
                        int(cols[p1ind]),
                        cols[s1ind],
                        cols[c2ind],
                        int(cols[p2ind]),
                        cols[s2ind],
                        cols[ptind],
                    )
            else:
                line_buffer.append(stripline)
                cols_buffer.append(cols)

                c1.append(fetchadd(cols[c1ind], chromDict))
                c2.append(fetchadd(cols[c2ind], chromDict))
                p1.append(int(cols[p1ind]))
                p2.append(int(cols[p2ind]))
                s1.append(fetchadd(cols[s1ind], strandDict))
                s2.append(fetchadd(cols[s2ind], strandDict))

        else:  # when everything is loaded in memory...

            res = _filterbycov(c1, p1, c2, p2, max_dist, method)

            for i in range(len(res)):
                # not high-frequency interactor pairs:
                if not res[i] > max_cov:
                    outstream.write(line_buffer[i])
                    # don't forget terminal newline
                    outstream.write("\n")
                    if out_stat:
                        out_stat.add_pair(
                            cols_buffer[i][c1ind],
                            int(cols_buffer[i][p1ind]),
                            cols_buffer[i][s1ind],
                            cols_buffer[i][c2ind],
                            int(cols_buffer[i][p2ind]),
                            cols_buffer[i][s2ind],
                            cols_buffer[i][ptind],
                        )
                # high-frequency interactor pairs:
                else:
                    if out_stat:
                        out_stat.add_pair(
                            cols_buffer[i][c1ind],
                            int(cols_buffer[i][p1ind]),
                            cols_buffer[i][s1ind],
                            cols_buffer[i][c2ind],
                            int(cols_buffer[i][p2ind]),
                            cols_buffer[i][s2ind],
                            "FF",
                        )
                    if outstream_high:
                        outstream_high.write(
                            # DD-marked pair:
                            sep.join(mark_split_pair_as_dup(cols_buffer[i]))
                            if mark_multi
                            # pair as is:
                            else line_buffer[i]
                        )
                        # don't forget terminal newline
                        outstream_high.write("\n")

            # flush buffers and perform necessary checks here:
            c1 = []
            c2 = []
            p1 = []
            p2 = []
            s1 = []
            s2 = []
            line_buffer = line_buffer[len(res) :]
            cols_buffer = cols_buffer[len(res) :]
            if not stripline:
                if len(line_buffer) != 0:
                    raise ValueError(
                        "{} lines left in the buffer, ".format(len(line_buffer))
                        + "should be none;"
                        + "something went terribly wrong"
                    )
                break

            break



import sys
import click

import random

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

UTIL_NAME = "pairtools_sample"


@cli.command()
@click.argument("fraction", type=float, required=True)
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@click.option(
    "-s",
    "--seed",
    type=int,
    default=None,
    help="the seed of the random number generator.",
)
@common_io_options
def sample(fraction, pairs_path, output, seed, **kwargs):
    """Select a random subset of pairs in a pairs file.

    FRACTION: the fraction of the randomly selected pairs subset

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.

    """
    sample_py(fraction, pairs_path, output, seed, **kwargs)


def sample_py(fraction, pairs_path, output, seed, **kwargs):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))

    random.seed(seed)

    for line in body_stream:
        if random.random() <= fraction:
            outstream.write(line)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    sample()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import sys
import click
import pandas as pd

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.scaling import compute_scaling


UTIL_NAME = "pairtools_scaling"


@cli.command()
@click.argument("input_path", type=str, nargs=-1, required=False)
@click.option(
    "-o", "--output", type=str, default="", help="output .tsv file with summary."
)
@click.option(
    "--view",
    "--regions",
    help="Path to a BED file which defines which regions (viewframe) of the chromosomes to use. "
    "By default, this is parsed from .pairs header. ",
    type=str,
    required=False,
    default=None,
)
@click.option(
    "--chunksize",
    type=int,
    default=100_000,
    show_default=True,
    required=False,
    help="Number of pairs in each chunk. Reduce for lower memory footprint.",
)
@click.option(
    "--dist-range",
    type=click.Tuple([int, int]),
    default=(10, 1_000_000_000),
    show_default=True,
    required=False,
    help="Distance range. ",
)
@click.option(
    "--n-dist-bins",
    type=int,
    default=128,
    show_default=True,
    required=False,
    help="Number of distance bins to split the distance range. ",
)
@common_io_options
def scaling(input_path, output, view, chunksize, dist_range, n_dist_bins, **kwargs):
    """Calculate pairs scalings.

    INPUT_PATH : by default, a .pairs/.pairsam file to calculate statistics.
    If not provided, the input is read from stdin.

    The files with paths ending with .gz/.lz4 are decompressed by bgzip/lz4c.

    Output is .tsv file with scaling stats (both cis scalings and trans levels).
    """
    scaling_py(input_path, output, view, chunksize, dist_range, n_dist_bins, **kwargs)


def scaling_py(input_path, output, view, chunksize, dist_range, n_dist_bins, **kwargs):

    if len(input_path) == 0:
        raise ValueError(f"No input paths: {input_path}")

    instream = fileio.auto_open(
        input_path[0],
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    if view is not None:
        view = pd.read_table(view)

    # Pass the header to the instream so that it can parse the header automatically
    cis_scalings, trans_levels = compute_scaling(
        instream,
        regions=view,
        chromsizes=None,
        dist_range=dist_range,
        n_dist_bins=n_dist_bins,
        chunksize=chunksize,
    )
    summary_stats = pd.concat([cis_scalings, trans_levels])

    # save statistics to the file
    summary_stats.to_csv(outstream, sep="\t")

    if instream != sys.stdin:
        instream.close()
    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    scaling()



"""
Legacy code:
``mark_duplicates`` is an offline method that finds duplicates in a given
input dataset.

For other applications on much larger datasets you may consider an online
method ``OnlineDuplicateDetector`` which is implemented as a class.
Note that for both methods data types are fixed:
    * chromosomes are int32
    * position is int32
    * strand is int32, which is basically the same as C type "char".
"""
import numpy as np
import cython

cimport numpy as np
cimport cython


### Online deduplicator used in pairtools.dedup Cython:
cdef class OnlineDuplicateDetector(object):
    cdef cython.int [:] c1
    cdef cython.int [:] c2
    cdef cython.int [:] p1
    cdef cython.int [:] p2
    cdef cython.int [:] s1
    cdef cython.int [:] s2
    cdef cython.char [:] rm
    cdef cython.int [:] parent_idxs
    cdef int methodid
    cdef int low
    cdef int high
    cdef int N
    cdef int max_mismatch
    cdef int returnData
    cdef int keep_parent_id

    def __init__(self, method, max_mismatch, returnData=False, keep_parent_id=False):
        if returnData == False:
            self.returnData = 0
        else:
            self.returnData = 1
        if keep_parent_id == False:
            self.keep_parent_id = 0
        else:
            self.keep_parent_id = 1
            self.parent_idxs = np.zeros(0, np.int32)

        self.N = 0
        self.c1 = np.zeros(0, np.int32)
        self.c2 = np.zeros(0, np.int32)
        self.p1 = np.zeros(0, np.int32)
        self.p2 = np.zeros(0, np.int32)
        self.s1 = np.zeros(0, np.int32)
        self.s2 = np.zeros(0, np.int32)

        self.rm = np.zeros(0, np.int8)
        if method == "max":
            self.methodid = 0
        elif method == "sum":
            self.methodid = 1
        else:
            raise ValueError('method should be "sum" or "max"')
        self.max_mismatch = int(max_mismatch)
        self.low = 0
        self.high = 1

    def _shrink(self):
        if self.returnData == 1:
            firstret = self.rm[:self.low]
            retainMask = (np.asarray(firstret) == False)
            del firstret
            ret = []
            for ar in [self.c1, self.c2, self.p1, self.p2, self.s1, self.s2]:
                ret.append(np.asarray(ar)[:self.low][retainMask])
        self.c1 = self.c1[self.low:]
        self.c2 = self.c2[self.low:]
        self.p1 = self.p1[self.low:]
        self.p2 = self.p2[self.low:]
        self.s1 = self.s1[self.low:]
        self.s2 = self.s2[self.low:]
        pastrm = self.rm[:self.low]
        self.rm = self.rm[self.low:]
        self.high = self.high-self.low
        self.N = self.N - self.low
        if self.returnData == 1:
            self.low = 0
            return ret
        if self.keep_parent_id == 1: # Return parent readIDs alongside with duplicates mask:
            pastidx = self.parent_idxs[:self.low]
            self.low = 0
            return pastrm, pastidx
        self.low = 0
        return pastrm

    def _run(self, finish=False):
        cdef int finishing = 0
        cdef int extraCondition

        if finish:
            finishing = 1

        while True:
            if self.low == self.N:
                break

            if self.high == self.N:
                if finishing == 1:
                    self.low += 1
                    self.high = self.low + 1
                    continue
                else:
                    break

            if self.rm[self.low] == 1:
                self.low += 1
                self.high = self.low+1
                continue

            # if high already removed, just continue
            if self.rm[self.high] == 1:
                self.high += 1
                continue

            # if we jumped too far, continue
            if ((self.c1[self.high] != self.c1[self.low]) or
                (self.p1[self.high] - self.p1[self.low] > self.max_mismatch)   or
                (self.p1[self.high] - self.p1[self.low] < 0  )):
                self.low += 1
                self.high = self.low + 1  # restart high
                continue

            if self.methodid == 0:
                extraCondition = max(
                    abs(self.p1[self.low] - self.p1[self.high]),
                    abs(self.p2[self.low] - self.p2[self.high])) <= self.max_mismatch
            elif self.methodid == 1:
                # sum of distances <= max_mismatch
                extraCondition = (
                    abs(self.p1[self.low] - self.p1[self.high]) +
                    abs(self.p2[self.low] - self.p2[self.high]) <= self.max_mismatch
                )
            else:
                raise ValueError(
                    "Unknown method id, this should not happen. "
                    "Check code of this function.")

            if ((self.c2[self.low] == self.c2[self.high]) and
                    (self.s1[self.low] == self.s1[self.high]) and
                    (self.s2[self.low] == self.s2[self.high]) and
                    extraCondition):
                self.rm[self.high] = 1
                if self.keep_parent_id == 1:
                    self.parent_idxs[self.high] = self.low
                self.high += 1
                continue
            self.high += 1

        return self._shrink()

    def push(self, c1, c2, p1, p2, s1, s2):
        self.c1 = np.concatenate([self.c1, c1])
        self.c2 = np.concatenate([self.c2, c2])
        self.p1 = np.concatenate([self.p1, p1])
        self.p2 = np.concatenate([self.p2, p2])
        self.s1 = np.concatenate([self.s1, s1])
        self.s2 = np.concatenate([self.s2, s2])
        self.rm = np.concatenate([self.rm, np.zeros(len(c1), dtype=np.int8)])
        if self.keep_parent_id == 1:
            self.parent_idxs = np.concatenate([self.parent_idxs, np.zeros(len(c1), dtype=np.int32)])
        self.N = self.N + len(c1)
        return self._run(finish=False)

    def finish(self):
        return self._run(finish=True)

    def getLen(self):
        return int(self.N)


# -*- coding: utf-8 -*-
from pairtools.lib import headerops

import pytest


def test_make_standard_header():
    header = headerops.make_standard_pairsheader()

    assert any([l.startswith("## pairs format") for l in header])
    assert any([l.startswith("#shape") for l in header])
    assert any([l.startswith("#columns") for l in header])

    header = headerops.make_standard_pairsheader(
        chromsizes=[("b", 100), ("c", 100), ("a", 100)]
    )

    assert sum([l.startswith("#chromsize") for l in header]) == 3


def test_samheaderops():
    header = headerops.make_standard_pairsheader()
    samheader = [
        "@SQ\tSN:chr1\tLN:100",
        "@SQ\tSN:chr2\tLN:100",
        "@SQ\tSN:chr3\tLN:100",
        "@PG\tID:bwa\tPN:bwa\tCL:bwa",
        "@PG\tID:bwa-2\tPN:bwa\tCL:bwa\tPP:bwa",
    ]
    header_with_sam = headerops.insert_samheader(header, samheader)

    assert len(header_with_sam) == len(header) + len(samheader)
    for l in samheader:
        assert any([l2.startswith("#samheader") and l in l2 for l2 in header_with_sam])

    # test adding new programs to the PG chain
    header_extra_pg = headerops.append_new_pg(header_with_sam, ID="test", PN="test")

    # test if all lines got transferred
    assert all([(old_l in header_extra_pg) for old_l in header_with_sam])
    # test if one PG got added
    assert len(header_extra_pg) == len(header_with_sam) + 1

    # test if the new PG has PP matching the ID of one of already existing PGs
    new_l = [l for l in header_extra_pg if l not in header_with_sam][0]
    pp = [f[3:] for f in new_l.split("\t") if f.startswith("PP:")][0]
    assert (
        len(
            [
                l
                for l in header_extra_pg
                if l.startswith("#samheader") and ("\tID:{}\t".format(pp) in l)
            ]
        )
        == 1
    )


def test_merge_pairheaders():
    headers = [["## pairs format v1.0"], ["## pairs format v1.0"]]
    merged_header = headerops._merge_pairheaders(headers)
    assert merged_header == headers[0]

    headers = [["## pairs format v1.0", "#a"], ["## pairs format v1.0", "#b"]]
    merged_header = headerops._merge_pairheaders(headers)
    assert merged_header == ["## pairs format v1.0", "#a", "#b"]

    headers = [
        ["## pairs format v1.0", "#chromsize: chr1 100", "#chromsize: chr2 200"],
        ["## pairs format v1.0", "#chromsize: chr1 100", "#chromsize: chr2 200"],
    ]
    merged_header = headerops._merge_pairheaders(headers)
    assert merged_header == headers[0]


def test_merge_different_pairheaders():
    with pytest.raises(Exception):
        headers = [["## pairs format v1.0"], ["## pairs format v1.1"]]
        merged_header = headerops._merge_pairheaders(headers)


def test_force_merge_pairheaders():
    headers = [
        ["## pairs format v1.0", "#chromsize: chr1 100"],
        ["## pairs format v1.0", "#chromsize: chr2 200"],
    ]
    merged_header = headerops._merge_pairheaders(headers, force=True)
    assert merged_header == [
        "## pairs format v1.0",
        "#chromsize: chr1 100",
        "#chromsize: chr2 200",
    ]


def test_merge_samheaders():
    headers = [
        ["@HD\tVN:1"],
        ["@HD\tVN:1"],
    ]
    merged_header = headerops._merge_samheaders(headers)
    assert merged_header == headers[0]

    headers = [
        [
            "@HD\tVN:1",
            "@SQ\tSN:chr1\tLN:100",
            "@SQ\tSN:chr2\tLN:100",
        ],
        [
            "@HD\tVN:1",
            "@SQ\tSN:chr1\tLN:100",
            "@SQ\tSN:chr2\tLN:100",
        ],
    ]
    merged_header = headerops._merge_samheaders(headers)
    assert merged_header == headers[0]

    headers = [
        [
            "@HD\tVN:1",
            "@PG\tID:bwa\tPN:bwa\tPP:cat",
        ],
        [
            "@HD\tVN:1",
            "@PG\tID:bwa\tPN:bwa\tPP:cat",
        ],
    ]
    merged_header = headerops._merge_samheaders(headers)
    print(merged_header)
    assert merged_header == [
        "@HD\tVN:1",
        "@PG\tID:bwa-1\tPN:bwa\tPP:cat-1",
        "@PG\tID:bwa-2\tPN:bwa\tPP:cat-2",
    ]


def test_merge_headers():
    headers = [
        [
            "## pairs format v1.0",
            "#samheader: @HD\tVN:1",
            "#samheader: @SQ\tSN:chr1\tLN:100",
            "#samheader: @SQ\tSN:chr2\tLN:100",
        ]
    ] * 2

    merged_header = headerops.merge_headers(headers)
    assert merged_header == headers[0]



"""
Set of functions used for pairsam parse, migrated from pairtools/pairtools_parse.py

Parse operates with several basic data types:

I. pysam-based:
    1. **sam entry** is a continuous aligned fragment of the read mapped to certain location in the genome.
        Because we read sam entries from .sam/.bam files automatically with modified pysam,
        each sam entry is in fact special AlignedSegmentPairtoolized Cython object
        that has alignment attributes and can be easily accessed from Python.
    
        Sam entries are gathered into reads by `push_pysam` function.
    
    2. **read** is a collection of sam entries corresponding to a single Hi-C molecule.
        It is represented by three variables:
        readID, sams1 and sams2, which keep left and right sam entries, correspondingly.
        Read is populated from the stream of sam entries on a fly, the process happenning
        in `streaming_classify` function.

II. python-based data types are parsed from pysam-based ones:

    1. **alignment** is a continuous aligned fragment represented as dictionary with relevant fields,
        such as "chrom", "pos5", "pos3", "strand", "type", etc.
        
        `empty_alignment` creates empty alignment,
        `parse_pysam_entry` create new alignments from pysam entries,
        `mask_alignment` clears some fields of the alignment to match the default "unmapped" state.
        
        `flip_alignment`, `flip_orientation` and `flip_ends` are useful functions that help to orient alignments.

    2. **pair** of two alignments is represented by three variables:
        algn1 (left alignment), algn2 (right alignment) and pair_index.
        Pairs are obtained by `parse_read` or `parse2_read`.
        Additionally, these functions also output all alignments for each side.

"""
from . import pairsam_format
from .parse_pysam import get_mismatches_c


def streaming_classify(
    instream, outstream, chromosomes, out_alignments_stream, out_stat, **kwargs
):
    """
    Parse input sam file into individual reads, pairs, walks,
    then write to the outstream(s).

    Additional kwargs:
        min_mapq,
        drop_readid,
        drop_seq,
        drop_sam,
        add_pair_index,
        add_columns, # comma-separated list
        report_alignment_end,
        max_inter_align_gap
    parse:
        max_molecule_size
        walks_policy
    parse2:
        single_end: indicator whether single-end data is provided
        report_position, one of: "outer", "junction", "read", "walk"
        report_orientation, one of: "pair", "junction", "read", "walk"
        dedup_max_mismatch: For intramolecular deduplication
        max_insert_size: maximum insert size when searching for overlapping ends of R1 and R2
        expand: perform combinatorial expansion or not
        max_expansion_depth: maximum expansion depth, works in combination with expand=True

    """

    parse2 = kwargs.get("parse2", False)

    ### Store output parameters in a usable form:
    chrom_enum = dict(
        zip(
            [pairsam_format.UNMAPPED_CHROM] + list(chromosomes),
            range(len(chromosomes) + 1),
        )
    )
    add_columns = kwargs.get("add_columns", "")
    if isinstance(add_columns, str) and len(add_columns) > 0:
        add_columns = add_columns.split(",")
    elif len(add_columns) == 0:
        add_columns = []
    elif not isinstance(add_columns, list):
        raise ValueError(f"Unknown type of add_columns: {type(add_columns)}")

    sam_tags = [col for col in add_columns if len(col) == 2 and col.isupper()]
    store_seq = "seq" in add_columns

    ### Compile readID transformation:
    readID_transform = kwargs.get("readid_transform", None)
    if readID_transform is not None:
        readID_transform = compile(readID_transform, "<string>", "eval")

    ### Prepare for iterative parsing of the input stream
    # Each read is represented by readID, sams1 (left alignments) and sams2 (right alignments)
    readID = ""  # Read id of the current read
    sams1 = []  # Placeholder for the left alignments
    sams2 = []  # Placeholder for the right alignments
    # Each read is comprised of multiple alignments, or sam entries:
    sam_entry = ""  # Placeholder for each aligned segment
    # Keep the id of the previous sam entry to detect when the read is completely populated:
    prev_readID = ""  # Placeholder for the read id

    ### Iterate over input pysam:
    instream = iter(instream)
    while sam_entry is not None:
        sam_entry = next(instream, None)

        readID = sam_entry.query_name if sam_entry else None
        if readID_transform is not None and readID is not None:
            readID = eval(readID_transform)

        # Read is fully populated, then parse and write:
        if not (sam_entry) or ((readID != prev_readID) and prev_readID):

            ### Parse
            if not parse2:  # regular parser:
                pairstream, all_algns1, all_algns2 = parse_read(
                    sams1,
                    sams2,
                    min_mapq=kwargs["min_mapq"],
                    max_molecule_size=kwargs["max_molecule_size"],
                    max_inter_align_gap=kwargs["max_inter_align_gap"],
                    walks_policy=kwargs["walks_policy"],
                    sam_tags=sam_tags,
                    store_seq=store_seq,
                    report_mismatches=True if "mismatches" in add_columns else False,
                )
            else:  # parse2 parser:
                pairstream, all_algns1, all_algns2 = parse2_read(
                    sams1,
                    sams2,
                    min_mapq=kwargs["min_mapq"],
                    max_inter_align_gap=kwargs["max_inter_align_gap"],
                    max_insert_size=kwargs.get("max_insert_size", 500),
                    single_end=kwargs["single_end"],
                    report_position=kwargs["report_position"],
                    report_orientation=kwargs["report_orientation"],
                    sam_tags=sam_tags,
                    dedup_max_mismatch=kwargs["dedup_max_mismatch"],
                    store_seq=store_seq,
                    expand=kwargs["expand"],
                    max_expansion_depth=kwargs["max_expansion_depth"],
                    report_mismatches=True if "mismatches" in add_columns else False,
                )

            ### Write:
            read_has_alignments = False
            for (algn1, algn2, pair_index) in pairstream:
                read_has_alignments = True

                # Alignment end defaults to 5' if report_alignment_end is unspecified:
                if kwargs.get("report_alignment_end", "5") == "5":
                    algn1["pos"] = algn1["pos5"]
                    algn2["pos"] = algn2["pos5"]
                else:
                    algn1["pos"] = algn1["pos3"]
                    algn2["pos"] = algn2["pos3"]

                if kwargs["flip"]:
                    flip_pair = not check_pair_order(algn1, algn2, chrom_enum)
                    if flip_pair:
                        algn1, algn2 = algn2, algn1
                        sams1, sams2 = sams2, sams1

                write_pairsam(
                    algn1,
                    algn2,
                    readID=prev_readID,
                    pair_index=pair_index,
                    sams1=sams1,
                    sams2=sams2,
                    out_file=outstream,
                    drop_readid=kwargs["drop_readid"],
                    drop_seq=kwargs["drop_seq"],
                    drop_sam=kwargs["drop_sam"],
                    add_pair_index=kwargs["add_pair_index"],
                    add_columns=add_columns,
                )

                # add a pair to PairCounter for stats output:
                if out_stat:
                    out_stat.add_pair(
                        algn1["chrom"],
                        int(algn1["pos"]),
                        algn1["strand"],
                        algn2["chrom"],
                        int(algn2["pos"]),
                        algn2["strand"],
                        algn1["type"] + algn2["type"],
                    )

            # write all alignments:
            if out_alignments_stream and read_has_alignments:
                write_all_algnments(
                    prev_readID, all_algns1, all_algns2, out_alignments_stream
                )

            # Empty read after writing:
            sams1.clear()
            sams2.clear()

        if sam_entry is not None:
            push_pysam(sam_entry, sams1, sams2)
            prev_readID = readID


############################
### Alignment utilities: ###
############################


def push_pysam(sam_entry, sams1, sams2):
    """Parse pysam AlignedSegment (sam) into pairtools sams entry"""
    flag = sam_entry.flag
    if (flag & 0x40) != 0:
        sams1.append(sam_entry)  # left read, or first read in a pair
    else:
        sams2.append(sam_entry)  # right read, or mate pair
    return


def empty_alignment():
    return {
        "chrom": pairsam_format.UNMAPPED_CHROM,
        "pos5": pairsam_format.UNMAPPED_POS,
        "pos3": pairsam_format.UNMAPPED_POS,
        "pos": pairsam_format.UNMAPPED_POS,
        "strand": pairsam_format.UNMAPPED_STRAND,
        "dist_to_5": 0,
        "dist_to_3": 0,
        "mapq": 0,
        "is_unique": False,
        "is_mapped": False,
        "is_linear": True,
        "cigar": "*",
        "algn_ref_span": 0,
        "algn_read_span": 0,
        "matched_bp": 0,
        "clip3_ref": 0,
        "clip5_ref": 0,
        "read_len": 0,
        "type": "N",
        "mismatches": "",
    }


def parse_pysam_entry(
    sam,
    min_mapq,
    sam_tags=None,
    store_seq=False,
    report_3_alignment_end=False,
    report_mismatches=False,
):
    """Parse alignments from pysam AlignedSegment entry
    :param sam: input pysam AlignedSegment entry
    :param min_mapq: minimal MAPQ to consider as a proper alignment
    :param sam_tags: list of sam tags to store
    :param store_seq: if True, the sequence will be parsed and stored in the output
    :param report_3_alignment_end: if True, 3'-end of alignment will be
                                reported as position (will be deprecated)
    :param report_mismatches: if True, mismatches will be parsed from MD field
    :return: parsed aligned entry (dictionary)
    """

    flag = sam.flag
    is_mapped = (flag & 0x04) == 0
    mapq = sam.mapq
    is_unique = sam.is_unique(min_mapq)
    is_linear = sam.is_linear
    cigar = sam.cigar_dict
    if is_mapped:
        if (flag & 0x10) == 0:
            strand = "+"
            dist_to_5 = cigar["clip5_ref"]
            dist_to_3 = cigar["clip3_ref"]
        else:
            strand = "-"
            dist_to_5 = cigar["clip3_ref"]
            dist_to_3 = cigar["clip5_ref"]

        if is_unique:
            chrom = sam.reference_name
            if strand == "+":
                # Note that pysam output is zero-based, thus add +1:
                pos5 = sam.reference_start + 1
                pos3 = sam.reference_start + cigar["algn_ref_span"]
            else:
                pos5 = sam.reference_start + cigar["algn_ref_span"]
                # Note that pysam output is zero-based, thus add +1:
                pos3 = sam.reference_start + 1

            # Get number of matches:
            if not sam.has_tag("MD") or not report_mismatches:
                mismatches = ""
            else:
                seq = sam.query_sequence.upper()
                quals = sam.query_qualities
                aligned_pairs = sam.get_aligned_pairs(with_seq=True, matches_only=True)
                mismatches = get_mismatches_c(seq, quals, aligned_pairs)
                mismatches = ",".join(
                    [
                        f"{original}:{mutated}:{phred}:{ref}:{read}"
                        for original, mutated, phred, ref, read in mismatches
                    ]
                )
                # n_matches = len(aligned_pairs)

        else:
            chrom = pairsam_format.UNMAPPED_CHROM
            strand = pairsam_format.UNMAPPED_STRAND
            pos5 = pairsam_format.UNMAPPED_POS
            pos3 = pairsam_format.UNMAPPED_POS
            mismatches = ""
    else:
        chrom = pairsam_format.UNMAPPED_CHROM
        strand = pairsam_format.UNMAPPED_STRAND
        pos5 = pairsam_format.UNMAPPED_POS
        pos3 = pairsam_format.UNMAPPED_POS

        dist_to_5 = 0
        dist_to_3 = 0
        mismatches = ""

    algn = {
        "chrom": chrom,
        "pos5": pos5,
        "pos3": pos3,
        "strand": strand,
        "mapq": mapq,
        "is_mapped": is_mapped,
        "is_unique": is_unique,
        "is_linear": is_linear,
        "dist_to_5": dist_to_5,
        "dist_to_3": dist_to_3,
        "type": ("N" if not is_mapped else ("M" if not is_unique else "U")),
        "mismatches": mismatches,
    }

    algn.update(cigar)

    algn["pos"] = algn["pos3"] if report_3_alignment_end else algn["pos5"]

    ### Add tags to the alignment:
    if sam_tags:
        tags = sam.tags
        for tag in sam_tags:
            algn[tag] = ""

        for col, value in tags:
            for tag in sam_tags:
                if col == tag:
                    algn[tag] = value
                    continue

    if store_seq:
        algn["seq"] = sam.seq

    return algn


def mask_alignment(algn):
    """
    Reset the coordinates of an alignment.
    """
    algn["chrom"] = pairsam_format.UNMAPPED_CHROM
    algn["pos5"] = pairsam_format.UNMAPPED_POS
    algn["pos3"] = pairsam_format.UNMAPPED_POS
    algn["pos"] = pairsam_format.UNMAPPED_POS
    algn["strand"] = pairsam_format.UNMAPPED_STRAND

    return algn


def flip_alignment(hic_algn):
    """
    Flip a single alignment as if it was sequenced from the opposite end
    :param hic_algn: Alignment to be modified
    :return:
    """
    hic_algn = dict(hic_algn)  # overwrite the variable with the copy of dictionary
    hic_algn["pos5"], hic_algn["pos3"] = hic_algn["pos3"], hic_algn["pos5"]
    hic_algn["strand"] = "+" if (hic_algn["strand"] == "-") else "-"
    return hic_algn


def flip_orientation(hic_algn):
    """
    Flip orientation of a single alignment
    :param hic_algn: Alignment to be modified
    :return:
    """
    hic_algn = dict(hic_algn)  # overwrite the variable with the copy of dictionary
    hic_algn["strand"] = "+" if (hic_algn["strand"] == "-") else "-"
    return hic_algn


def flip_position(hic_algn):
    """
    Flip ends of a single alignment
    :param hic_algn: Alignment to be modified
    :return:
    """
    hic_algn = dict(hic_algn)  # overwrite the variable with the copy of dictionary
    hic_algn["pos5"], hic_algn["pos3"] = hic_algn["pos3"], hic_algn["pos5"]
    return hic_algn


####################
### Parsing utilities:
####################


def parse_read(
    sams1,
    sams2,
    min_mapq,
    max_molecule_size,
    max_inter_align_gap,
    walks_policy,
    sam_tags,
    store_seq,
    report_mismatches=False,
):
    """
    Parse sam entries corresponding to a single read (or Hi-C molecule)
    into pairs of alignments.

    Returns
    -------
    stream: iterator
        Each element is a triplet: (algn1, aldn2, pair_index)
        algn1, algn2: dict
            Two alignments selected for reporting as a Hi-C pair.
        pair_index
            pair index of a pair in the molecule.
    algns1, algns2: lists
        All alignments, sorted according to their order in on a read.
    """

    # Check if there is at least one sam entry per side:
    if walks_policy == "all":
        is_empty = (len(sams1) == 0 and len(sams2) < 2) or (
            len(sams2) == 0 and len(sams1) < 2
        )
    else:
        is_empty = (len(sams1) == 0) or (len(sams2) == 0)
    if is_empty:
        algns1 = [empty_alignment()]
        algns2 = [empty_alignment()]
        algns1[0]["type"] = "X"
        algns2[0]["type"] = "X"
        pair_index = (1, "R1-2")
        return iter([(algns1[0], algns2[0], pair_index)]), algns1, algns2

    # Generate a sorted, gap-filled list of all alignments
    algns1 = [
        parse_pysam_entry(
            sam, min_mapq, sam_tags, store_seq, report_mismatches=report_mismatches
        )
        for sam in sams1
    ]
    algns2 = [
        parse_pysam_entry(
            sam, min_mapq, sam_tags, store_seq, report_mismatches=report_mismatches
        )
        for sam in sams2
    ]

    if len(algns1) > 0:
        algns1 = sorted(algns1, key=lambda algn: algn["dist_to_5"])
    else:
        algns1 = [empty_alignment()]  # Empty alignment dummy
    if len(algns2) > 0:
        algns2 = sorted(algns2, key=lambda algn: algn["dist_to_5"])
    else:
        algns2 = [empty_alignment()]  # Empty alignment dummy

    if max_inter_align_gap is not None:
        _convert_gaps_into_alignments(algns1, max_inter_align_gap)
        _convert_gaps_into_alignments(algns2, max_inter_align_gap)

    # By default, assume each molecule is a single pair with single unconfirmed pair:
    hic_algn1 = algns1[0]
    hic_algn2 = algns2[0]
    pair_index = (1, "R1-2")

    # Define the type of alignment on each side:
    is_chimeric_1 = len(algns1) > 1
    is_chimeric_2 = len(algns2) > 1

    # Parse chimeras
    if is_chimeric_1 or is_chimeric_2:

        # Report all the linear alignments in a read pair
        if walks_policy == "all":
            # Report linear alignments after deduplication of complex walks with default settings:
            return (
                parse_complex_walk(
                    algns1,
                    algns2,
                    max_molecule_size,
                    report_position="outer",
                    report_orientation="pair",
                ),
                algns1,
                algns2,
            )

        elif walks_policy in ["mask", "5any", "5unique", "3any", "3unique"]:
            # Report only two alignments for a read pair
            rescued_linear_side = rescue_walk(algns1, algns2, max_molecule_size)

            # Walk was rescued as a simple walk:
            if rescued_linear_side is not None:
                pair_index = (1, "R1" if rescued_linear_side == 1 else "R2")
            # Walk is unrescuable:
            else:
                if walks_policy == "mask":
                    hic_algn1 = mask_alignment(dict(hic_algn1))
                    hic_algn2 = mask_alignment(dict(hic_algn2))
                    hic_algn1["type"] = "W"
                    hic_algn2["type"] = "W"

                elif walks_policy == "5any":
                    hic_algn1 = algns1[0]
                    hic_algn2 = algns2[0]

                elif walks_policy == "5unique":
                    hic_algn1 = algns1[0]
                    for algn in algns1:
                        if algn["is_mapped"] and algn["is_unique"]:
                            hic_algn1 = algn
                            break

                    hic_algn2 = algns2[0]
                    for algn in algns2:
                        if algn["is_mapped"] and algn["is_unique"]:
                            hic_algn2 = algn
                            break

                elif walks_policy == "3any":
                    hic_algn1 = algns1[-1]
                    hic_algn2 = algns2[-1]

                elif walks_policy == "3unique":
                    hic_algn1 = algns1[-1]
                    for algn in algns1[::-1]:
                        if algn["is_mapped"] and algn["is_unique"]:
                            hic_algn1 = algn
                            break

                    hic_algn2 = algns2[-1]
                    for algn in algns2[::-1]:
                        if algn["is_mapped"] and algn["is_unique"]:
                            hic_algn2 = algn
                            break

                # lower-case reported walks on the chimeric side
                if walks_policy != "mask":
                    if is_chimeric_1:
                        hic_algn1 = dict(hic_algn1)
                        hic_algn1["type"] = hic_algn1["type"].lower()
                    if is_chimeric_2:
                        hic_algn2 = dict(hic_algn2)
                        hic_algn2["type"] = hic_algn2["type"].lower()

        else:
            raise ValueError(f"Walks policy {walks_policy} is not supported.")

    return iter([(hic_algn1, hic_algn2, pair_index)]), algns1, algns2


def parse2_read(
    sams1,
    sams2,
    min_mapq,
    max_inter_align_gap,
    max_insert_size,
    single_end,
    report_position="outer",
    report_orientation="pair",
    sam_tags=[],
    dedup_max_mismatch=3,
    store_seq=False,
    report_mismatches=False,
    expand=False,
    max_expansion_depth=None,
):
    """
    Parse sam entries corresponding to a Hi-C molecule into alignments in parse2 mode
    for a Hi-C pair.
    Returns
    -------
    stream: iterator
        Each element is a triplet: (algn1, aldn2, pair_index)
        algn1, algn2: dict
            Two alignments selected for reporting as a Hi-C pair.
        pair_index
            pair index of a pair in the molecule, a tuple: (1, "R1-2")
    algns1, algns2: lists
        All alignments, sorted according to their order in on a read.
    """

    # Single-end mode:
    if single_end:
        # Generate a sorted, gap-filled list of all alignments
        algns1 = [
            parse_pysam_entry(
                sam, min_mapq, sam_tags, store_seq, report_mismatches=report_mismatches
            )
            for sam in sams2  # note sams2, that's how these reads are typically parsed
        ]
        algns1 = sorted(algns1, key=lambda algn: algn["dist_to_5"])
        if max_inter_align_gap is not None:
            _convert_gaps_into_alignments(algns1, max_inter_align_gap)

        algns2 = [empty_alignment()]  # Empty alignment dummy

        if len(algns1) > 1:
            # Look for ligation pair, and report linear alignments after deduplication of complex walks:
            # (Note that coordinate system for single-end reads does not change the behavior)
            output = parse_complex_walk(
                algns1,
                algns2,
                max_insert_size,
                report_position,
                report_orientation,
                dedup_max_mismatch,
            )
            output = [x for x in output if x[-1][-1] != "R1-2"]
            return (output, algns1, algns2)
        elif len(algns1) == 1:
            # If no additional information, we assume each molecule is a single ligation with single unconfirmed pair:
            algn2 = empty_alignment()
            pair_index = (1, "R1")
            return iter([(algns1[0], algn2, pair_index)]), algns1, algns2
        else:
            # If no additional information, we assume each molecule is a single ligation with single unconfirmed pair:
            algn1 = empty_alignment()
            algn2 = empty_alignment()
            pair_index = (1, "R1")
            return iter([(algn1, algn2, pair_index)]), algns1, algns2

    # Paired-end mode:
    else:
        # Check if there is at least one SAM entry per side:
        is_empty = (len(sams1) == 0 and len(sams2) < 2) or (
            len(sams2) == 0 and len(sams1) < 2
        )
        if is_empty:
            algns1 = [empty_alignment()]
            algns2 = [empty_alignment()]
            algns1[0]["type"] = "X"
            algns2[0]["type"] = "X"
            pair_index = (1, "R1-2")
            return iter([(algns1[0], algns2[0], pair_index)]), algns1, algns2

        # Generate a sorted, gap-filled list of all alignments
        algns1 = [
            parse_pysam_entry(
                sam, min_mapq, sam_tags, store_seq, report_mismatches=report_mismatches
            )
            for sam in sams1
        ]
        algns2 = [
            parse_pysam_entry(
                sam, min_mapq, sam_tags, store_seq, report_mismatches=report_mismatches
            )
            for sam in sams2
        ]

        # Sort alignments by the distance to the 5'-end:
        if len(algns1) > 0:
            algns1 = sorted(algns1, key=lambda algn: algn["dist_to_5"])
        else:
            algns1 = [empty_alignment()]  # Empty alignment dummy
        if len(algns2) > 0:
            algns2 = sorted(algns2, key=lambda algn: algn["dist_to_5"])
        else:
            algns2 = [empty_alignment()]  # Empty alignment dummy

        # Convert alignment gaps to alignments:
        if max_inter_align_gap is not None:
            _convert_gaps_into_alignments(algns1, max_inter_align_gap)
            _convert_gaps_into_alignments(algns2, max_inter_align_gap)

        is_chimeric_1 = len(algns1) > 1
        is_chimeric_2 = len(algns2) > 1

        if is_chimeric_1 or is_chimeric_2:
            # If at least one side is chimera, we must look for ligation pair, and
            # report linear alignments after deduplication of complex walks:
            return (
                parse_complex_walk(
                    algns1,
                    algns2,
                    max_insert_size,
                    report_position,
                    report_orientation,
                    dedup_max_mismatch,
                    expand,
                    max_expansion_depth,
                ),
                algns1,
                algns2,
            )
        else:
            # If no additional information, we assume each molecule is a single ligation with single unconfirmed pair:
            algn2 = algns2[0]
            if report_orientation == "walk":
                algn2 = flip_orientation(algn2)
            if report_position == "walk":
                algn2 = flip_position(algn2)
            pair_index = (1, "R1-2")
            return iter([(algns1[0], algn2, pair_index)]), algns1, algns2


####################
### Walks utilities:
####################


def rescue_walk(algns1, algns2, max_molecule_size):
    """
    Rescue a single ligation that appears as a walk.
    Checks if a molecule with three alignments could be formed via a single
    ligation between two fragments, where one fragment was so long that it
    got sequenced on both sides.
    Uses three criteria:
    1) the 3'-end alignment on one side maps to the same chromosome as the
    alignment fully covering the other side (i.e. the linear alignment)
    2) the two alignments point towards each other on the chromosome
    3) the distance between the outer ends of the two alignments is below
    the specified threshold.
    Alternatively, a single ligation get rescued when the 3' sub-alignment
    maps to multiple locations or no locations at all.

    In the case of a successful rescue, tags the 3' sub-alignment with
    type='X' and the linear alignment on the other side with type='R'.
    Returns
    -------
    linear_side : int
        If the case of a successful rescue, returns the index of the side
        with a linear alignment.
    """

    # If both sides have one alignment or none, no need to rescue!
    n_algns1 = len(algns1)
    n_algns2 = len(algns2)

    if (n_algns1 <= 1) and (n_algns2 <= 1):
        return None

    # Can rescue only pairs with one chimeric alignment with two parts.
    if not (
        ((n_algns1 == 1) and (n_algns2 == 2)) or ((n_algns1 == 2) and (n_algns2 == 1))
    ):
        return None

    first_read_is_chimeric = n_algns1 > 1
    chim5_algn = algns1[0] if first_read_is_chimeric else algns2[0]
    chim3_algn = algns1[1] if first_read_is_chimeric else algns2[1]
    linear_algn = algns2[0] if first_read_is_chimeric else algns1[0]

    # the linear alignment must be uniquely mapped
    if not (linear_algn["is_mapped"] and linear_algn["is_unique"]):
        return None

    can_rescue = True
    # we automatically rescue chimeric alignments with null and non-unique
    # alignments at the 3' side
    if chim3_algn["is_mapped"] and chim5_algn["is_unique"]:
        # 1) in rescued walks, the 3' alignment of the chimeric alignment must be on
        # the same chromosome as the linear alignment on the opposite side of the
        # molecule
        can_rescue &= chim3_algn["chrom"] == linear_algn["chrom"]

        # 2) in rescued walks, the 3' supplemental alignment of the chimeric
        # alignment and the linear alignment on the opposite side must point
        # towards each other
        can_rescue &= chim3_algn["strand"] != linear_algn["strand"]
        if linear_algn["strand"] == "+":
            can_rescue &= linear_algn["pos5"] < chim3_algn["pos5"]
        else:
            can_rescue &= linear_algn["pos5"] > chim3_algn["pos5"]

        # 3) in single ligations appearing as walks, we can infer the size of
        # the molecule and this size must be smaller than the maximal size of
        # Hi-C molecules after the size selection step of the Hi-C protocol
        if linear_algn["strand"] == "+":
            molecule_size = (
                chim3_algn["pos5"]
                - linear_algn["pos5"]
                + chim3_algn["dist_to_5"]
                + linear_algn["dist_to_5"]
            )
        else:
            molecule_size = (
                linear_algn["pos5"]
                - chim3_algn["pos5"]
                + chim3_algn["dist_to_5"]
                + linear_algn["dist_to_5"]
            )

        can_rescue &= molecule_size <= max_molecule_size

    if can_rescue:
        # changing the type of the 3' alignment on side 1, does not show up in the output:
        if first_read_is_chimeric:

            algns1[1]["type"] = "X"
            algns2[0]["type"] = "R"
            return 1
        # changing the type of the 3' alignment on side 2, does not show up in the output:
        else:
            algns1[0]["type"] = "R"
            algns2[1]["type"] = "X"
            return 2
    else:
        return None


def _convert_gaps_into_alignments(sorted_algns, max_inter_align_gap):
    """
    Inplace conversion of gaps longer than max_inter_align_gap into alignments
    """
    if (len(sorted_algns) == 1) and (not sorted_algns[0]["is_mapped"]):
        return

    last_5_pos = 0
    for i in range(len(sorted_algns)):
        algn = sorted_algns[i]
        if algn["dist_to_5"] - last_5_pos > max_inter_align_gap:
            new_algn = empty_alignment()
            new_algn["dist_to_5"] = last_5_pos
            new_algn["algn_read_span"] = algn["dist_to_5"] - last_5_pos
            new_algn["read_len"] = algn["read_len"]
            new_algn["dist_to_3"] = new_algn["read_len"] - algn["dist_to_5"]

            last_5_pos = algn["dist_to_5"] + algn["algn_read_span"]

            sorted_algns.insert(i, new_algn)
            i += 2
        else:
            last_5_pos = max(last_5_pos, algn["dist_to_5"] + algn["algn_read_span"])
            i += 1


def parse_complex_walk(
    algns1,
    algns2,
    max_insert_size,
    report_position,
    report_orientation,
    dedup_max_mismatch=3,
    expand=False,
    max_expansion_depth=None,
):
    """
    Parse a set of ligations that appear as a complex walk.
    This procedure is equivalent to intramolecular deduplication that preserved pair order in a walk.

    :param algns1: List of sequential lefts alignments
    :param algns2: List of sequential right alignments
    :param max_insert_size: maximum insert size when searching for overlapping ends of R1 and R2
    :param report_position: one of "outer", "junction", "read", "walk"; sets pos5 and pos3
    :param report_orientation: one of "pair", "junction", "read", "walk"; sets strand
    :param dedup_max_mismatch: allowed mismatch between intramolecular alignments to detect readthrough duplicates
    :param expand: perform combinatorial expansion of pairs or not
    :param max_expansion_depth: maximum depth (number of segments separating pair). All by default.

    :return: iterator with parsed pairs

    **Intramolecular deduplication**

     Forward read (left):                       right read (right):
    5'------------------------->3'     3'<--------------------------5'
             algns1                              algns2
    <5---3><5---3><5---3><5---3>        <3---5><3---5><3---5><3---5>
       l0     l1    l2     l3              r3     r2     r1    r0

    Alignment - bwa mem reported hit or alignment after gaps conversion.
    Left and right alignments (algns1: [l0, l1, l2, l3], algns2: [r0, r1, r2, r3])
    - alignments on left and right reads reported from 5' to 3' orientation.

    Intramolecular deduplication consists of two steps:
    I. iterative search of overlapping alignment pairs (aka overlap),
    II. if no overlaps or search not possible (less than 2 alignments on either sides),
    search for overlap of end alignments (aka partial overlap).
    III. report pairs before the overlap, deduplicated pairs of overlap and pairs after that.

    Iterative search of overlap is in fact scanning of the right read pairs for the hit
    with the 3'-most pair of the left read:
        1. Initialize.
            Start from 3' of left and right reads. Set `current_left_pair` and `current_right_pair` pointers
        2. Initial compare.
            Compare pairs l2-l3 and r3-r2 by `pairs_overlap`.
                If successful, we found the overlap, go to reporting.
                If unsuccessful, continue search.
        3. Increment.
            Shift `current_right_pair` pointer by one (e.g., take the pair r2-r1).
        4. Check.
            Check that this pair can form a potential overlap with left alignments:
            the number of pairs downstream from l2-l3 on left read should not be less than
            the number of pairs upstream from r2-r1 on right read.
                If overlap cannot be formed, no other overlap in this complex walk is possible, safely exit.
                If the potential overlap can be formed, continue comparison.
        5. Compare.
            Compare the current pair of pairs on left and right reads.
                If comparison fails, go to step 3.
                If comparison is successful, go to 6.
        6. Verify.
            Check that downstream pairs on the left read overlap with the upstream pairs on the right read.
                 If yes, exit.
                 If not, we do not have an overlap, go to step 3.
    """

    AVAILABLE_REPORT_POSITION = ["outer", "junction", "read", "walk"]
    assert report_position in AVAILABLE_REPORT_POSITION, (
        f"Cannot report position {report_position}, as it is not implemented"
        f'Available choices are: {", ".join(AVAILABLE_REPORT_POSITION)}'
    )

    AVAILABLE_REPORT_ORIENTATION = ["pair", "junction", "read", "walk"]
    assert report_orientation in AVAILABLE_REPORT_ORIENTATION, (
        f"Cannot report orientation {report_orientation}, as it is not implemented"
        f'Available choices are: {", ".join(AVAILABLE_REPORT_ORIENTATION)}'
    )

    output_pairs = []

    # Initialize (step 1).
    n_algns1 = len(algns1)
    n_algns2 = len(algns2)
    current_left_pair = current_right_pair = 1
    remaining_left_pairs = (
        n_algns1 - 1
    )  # Number of possible pairs remaining on left read
    remaining_right_pairs = (
        n_algns2 - 1
    )  # Number of possible pairs remaining on right read
    checked_right_pairs = (
        0  # Number of checked pairs on right read (from the end of read)
    )
    is_overlap = False

    # I. Iterative search of overlap, at least two alignments on each side:
    if (n_algns1 >= 2) and (n_algns2 >= 2):
        # Iteration includes check (step 4):
        while (remaining_left_pairs > checked_right_pairs) and (
            remaining_right_pairs > 0
        ):
            pair1 = (algns1[-current_left_pair - 1], algns1[-current_left_pair])
            pair2 = (algns2[-current_right_pair - 1], algns2[-current_right_pair])
            # Compare (initial or not, step 2 or 5):
            is_overlap = pairs_overlap(
                pair1, pair2, dedup_max_mismatch=dedup_max_mismatch
            )
            if is_overlap:
                last_idx_left_temp = current_left_pair
                last_idx_right_temp = current_right_pair
                checked_right_temp = checked_right_pairs
                # Verify (step 6):
                while is_overlap and (checked_right_temp > 0):
                    last_idx_left_temp += 1
                    last_idx_right_temp -= 1
                    pair1 = (
                        algns1[-last_idx_left_temp - 1],
                        algns1[-last_idx_left_temp],
                    )
                    pair2 = (
                        algns2[-last_idx_right_temp - 1],
                        algns2[-last_idx_right_temp],
                    )
                    is_overlap &= pairs_overlap(
                        pair1, pair2, dedup_max_mismatch=dedup_max_mismatch
                    )
                    checked_right_temp -= 1
                if is_overlap:  # exit
                    current_right_pair += 1
                    break

            # Increment pointers (step 3)
            current_right_pair += 1
            checked_right_pairs += 1
            remaining_right_pairs -= 1

        # No overlap found, roll the current_idx_right back to the initial value:
        if not is_overlap:
            current_right_pair = 1

    # II. Search of partial overlap if there are less than 2 alignments at either sides, or no overlaps found
    if current_right_pair == 1:
        last_reported_alignment_left = last_reported_alignment_right = 1
        if partial_overlap(
            algns1[-1],
            algns2[-1],
            max_insert_size=max_insert_size,
            dedup_max_mismatch=dedup_max_mismatch,
        ):
            if (
                n_algns1 >= 2
            ):  # single alignment on right read and multiple alignments on left
                pair_index = (len(algns1) - 1, "R1")
                output_pairs.append(
                    format_pair(
                        algns1[-2],
                        algns1[-1],
                        pair_index=pair_index,
                        algn2_pos3=algns2[-1]["pos5"],
                        report_position=report_position,
                        report_orientation=report_orientation,
                    )
                )
                last_reported_alignment_left = 2  # set the pointer for reporting

            if (
                n_algns2 >= 2
            ):  # single alignment on left read and multiple alignments on right
                pair_index = (len(algns1), "R2")
                output_pairs.append(
                    format_pair(
                        algns2[-1],
                        algns2[-2],
                        pair_index=pair_index,
                        algn1_pos3=algns1[-1]["pos5"],
                        report_position=report_position,
                        report_orientation=report_orientation,
                    )
                )
                last_reported_alignment_right = 2  # set the pointer for reporting

            # Note that if n_algns1==n_algns2==1 and alignments overlap, then we don't need to check,
            # it's a non-ligated DNA fragment that we don't report.

        else:  # end alignments do not overlap, report regular pair:
            pair_index = (len(algns1), "R1-2")
            output_pairs.append(
                format_pair(
                    algns1[-1],
                    algns2[-1],
                    pair_index=pair_index,
                    report_position=report_position,
                    report_orientation=report_orientation,
                )
            )

    else:  # there was an overlap, set some pointers:
        last_reported_alignment_left = (
            last_reported_alignment_right
        ) = current_right_pair

    # III. Report all remaining alignments.
    # Report all unique alignments on left read (sequential):
    for i in range(0, n_algns1 - last_reported_alignment_left):
        pair_index = (i + 1, "R1")
        output_pairs.append(
            format_pair(
                algns1[i],
                algns1[i + 1],
                pair_index=pair_index,
                report_position=report_position,
                report_orientation=report_orientation,
            )
        )

    # Report the pairs where both left alignments overlap right:
    for i_overlapping in range(current_right_pair - 1):
        idx_left = n_algns1 - current_right_pair + i_overlapping
        idx_right = n_algns2 - 1 - i_overlapping
        pair_index = (idx_left + 1, "R1&2")
        output_pairs.append(
            format_pair(
                algns1[idx_left],
                algns1[idx_left + 1],
                pair_index=pair_index,
                algn2_pos3=algns2[idx_right - 1]["pos5"],
                report_position=report_position,
                report_orientation=report_orientation,
            )
        )

    # Report all the sequential chimeric pairs in the right read, but not the overlap:
    reporting_order = range(
        0, min(current_right_pair, n_algns2 - last_reported_alignment_right)
    )
    for i in reporting_order:
        # Determine the pair index depending on what is the overlap:
        shift = -1 if current_right_pair > 1 else 0
        pair_index = (
            (
                n_algns1
                + min(current_right_pair, n_algns2 - last_reported_alignment_right)
                - i
                + shift
            ),
            "R2",
        )
        output_pairs.append(
            format_pair(
                algns2[i + 1],
                algns2[i],
                pair_index=pair_index,
                report_position=report_position,
                report_orientation=report_orientation,
            )
        )

    # Sort the pairs according to the pair index:
    output_pairs.sort(key=lambda x: int(x[-1][0]))
    if expand:
        output_pairs = expand_pairs(output_pairs, max_expansion_depth)
    return iter(output_pairs)


### Additional functions for pairs ###
def expand_pairs(pairs_list, max_expansion_depth=None):
    """
    Perform combinatorial expansion of the pairs.

    Parameters
    ----------
    pairs_list: List of formatted pairs (triplets: algn1, algn2, pair_index).
    max_expansion_depth: maximum depth of expansion; all by default (None),
                    number will enforce only pairs from the same strand.
    Returns
    -------
    list of expanded pairs

    """

    for algn1, _algn1, pair_index1 in pairs_list:
        for _algn2, algn2, pair_index2 in pairs_list:
            if pair_index1 > pair_index2:
                continue
            elif pair_index1 == pair_index2:
                # output regular pair with no change
                yield algn1, _algn1, pair_index1
            else:
                pair_order1, pair_type1 = pair_index1
                pair_order2, pair_type2 = pair_index2
                separated_by = pair_order2 - pair_order1
                if (
                    pair_type1 == "R1-2"
                    or pair_type2 == "R1-2"
                    or (pair_type1 == "R1" and pair_type2 == "R2")
                ):
                    pair_type = "R1-2"
                elif pair_type1 == pair_type2:
                    pair_type = pair_type1
                elif pair_type1 == "R1&2":
                    pair_type = pair_type2
                elif pair_type2 == "R1&2":
                    pair_type = pair_type1
                else:
                    raise ValueError(
                        f"Unexpected error, pair types: {pair_type1}, {pair_type2}"
                    )
                same_read = pair_type != "R1-2"

                if (max_expansion_depth is None) or (
                    (separated_by <= max_expansion_depth) and same_read
                ):
                    pair_type = f"E{separated_by}_{pair_type}"
                    yield algn1, algn2, (pair_order1, pair_type)


### Additional functions for complex walks rescue ###
def partial_overlap(algn1, algn2, max_insert_size=500, dedup_max_mismatch=5):
    """
    Two ends of alignments overlap if:
     1) they are from the same chromosome,
     2) map in the opposite directions,
     3) the distance between the outer ends of the two alignments is below the specified max_insert_size,
     4) the distance between the outer ends of the two alignments is above the maximum alignment size.
    (4) guarantees that the alignments point towards each other on the chromosomes.

    Allowed mismatch between intramolecular alignments to detect readthrough duplicates.

    Return: 1 if the alignments overlap or both have troubles with unique mapping,
            0 if they do not overlap or if we do not have enough information
            (e.g. only one of the alignments have troubles with being mapped).
    """

    # Alignments with no match or with multiple matches are counted as overlaps
    if not (algn1["is_mapped"] and algn1["is_unique"]):
        if not (algn2["is_mapped"] and algn2["is_unique"]):
            return 1

    # We assume that successful alignment cannot be an overlap with unmapped or multi-mapped region
    if not (algn1["is_mapped"] and algn1["is_unique"]):
        return 0
    if not (algn2["is_mapped"] and algn2["is_unique"]):
        return 0

    # Both alignments are mapped and unique
    do_overlap = True

    do_overlap &= algn1["chrom"] == algn2["chrom"]
    do_overlap &= algn1["strand"] != algn2["strand"]

    if algn1["strand"] == "+":
        min_algn_size = max(
            algn1["pos3"] - algn1["pos5"], algn2["pos5"] - algn2["pos3"]
        )
        distance_outer_ends = algn2["pos5"] - algn1["pos5"]
    else:
        min_algn_size = max(
            algn1["pos5"] - algn1["pos3"], algn2["pos3"] - algn2["pos5"]
        )
        distance_outer_ends = algn1["pos5"] - algn2["pos5"]

    do_overlap &= distance_outer_ends <= max_insert_size + dedup_max_mismatch
    do_overlap &= distance_outer_ends >= min_algn_size - dedup_max_mismatch

    if do_overlap:
        return 1
    return 0


def pairs_overlap(algns1, algns2, dedup_max_mismatch=3):
    """
    We assume algns1 originate from left read, and algns2 originate from right read:
    left read:                             right read:
    ---------------------------->     <----------------------------
                algns1                             algns2
    5------------3_5------------3     3------------5_3------------5'
    left_5'-algn    left_3'-algn      right_3'-algn   right_5'-algn

    Two pairs of alignments overlap if:
    1) chromosomes/mapping/strand of left_5'-algn and right_3'-algn are the same,
    2) chromosomes/mapping/strand of left_3'-algn and right_5'-algn are the same,
    3) pos3 of left_5'-algn is close to pos5 of right_3'-algn (with dedup_max_mismatch), and
    4) pos5 of left_3'-algn is close to pos3 of right_5'-algn.

    Return: 1 of the pairs of alignments overlap, 0 otherwise.
    """
    left5_algn = algns1[0]
    left3_algn = algns1[1]
    right5_algn = algns2[0]
    right3_algn = algns2[1]

    # We assume that successful alignment cannot be an overlap with unmapped or multi-mapped region:
    mapped_left5_algn = left5_algn["is_mapped"] and left5_algn["is_unique"]
    mapped_left3_algn = left3_algn["is_mapped"] and left3_algn["is_unique"]
    mapped_right5_algn = right5_algn["is_mapped"] and right5_algn["is_unique"]
    mapped_right3_algn = right3_algn["is_mapped"] and right3_algn["is_unique"]

    if not mapped_left5_algn and not mapped_right3_algn:
        left_overlap = True
    elif not mapped_left5_algn and mapped_right3_algn:
        left_overlap = False
    elif mapped_left5_algn and not mapped_right3_algn:
        left_overlap = False
    else:
        left_overlap = True
        left_overlap &= left5_algn["chrom"] == right3_algn["chrom"]
        left_overlap &= left5_algn["strand"] != right3_algn["strand"]

    if not mapped_left3_algn and not mapped_right5_algn:
        right_overlap = True
    elif not mapped_left3_algn and mapped_right5_algn:
        right_overlap = False
    elif mapped_left3_algn and not mapped_right5_algn:
        right_overlap = False
    else:
        right_overlap = True
        right_overlap &= left3_algn["chrom"] == right5_algn["chrom"]
        right_overlap &= left3_algn["strand"] != right5_algn["strand"]

    same_pair = True
    same_pair &= abs(left5_algn["pos3"] - right3_algn["pos5"]) <= dedup_max_mismatch
    same_pair &= abs(left3_algn["pos5"] - right5_algn["pos3"]) <= dedup_max_mismatch

    if left_overlap & right_overlap & same_pair:
        return 1
    else:
        return 0


def format_pair(
    hic_algn1,
    hic_algn2,
    pair_index,
    report_position="outer",
    report_orientation="pair",
    algn1_pos5=None,
    algn1_pos3=None,
    algn2_pos5=None,
    algn2_pos3=None,
):
    """
    Return a triplet: pair of formatted alignments and pair_index in a walk

    :param hic_algn1: Left alignment forming a pair
    :param hic_algn2: Right alignment forming a pair
    :param algns1: All left read alignments for formal reporting
    :param algns2: All right read alignments for formal reporting
    :param pair_index: Index of the pair
    :param algn1_pos5: Replace reported 5'-position of the alignment 1 with this value
    :param algn1_pos3: Replace reported 3'-position of the alignment 1 with this value
    :param algn2_pos5: Replace reported 5'-position of the alignment 2 with this value
    :param algn2_pos3: Replace reported 3'-position of the alignment 2 with this value

    """
    # Make sure the original data is not modified:
    hic_algn1, hic_algn2 = dict(hic_algn1), dict(hic_algn2)

    # Adjust the 5' and 3'-ends:
    hic_algn1["pos5"] = algn1_pos5 if not algn1_pos5 is None else hic_algn1["pos5"]
    hic_algn1["pos3"] = algn1_pos3 if not algn1_pos3 is None else hic_algn1["pos3"]
    hic_algn2["pos5"] = algn2_pos5 if not algn2_pos5 is None else hic_algn2["pos5"]
    hic_algn2["pos3"] = algn2_pos3 if not algn2_pos3 is None else hic_algn2["pos3"]

    hic_algn1["type"] = (
        "N"
        if not hic_algn1["is_mapped"]
        else "M"
        if not hic_algn1["is_unique"]
        else "U"
    )

    hic_algn2["type"] = (
        "N"
        if not hic_algn2["is_mapped"]
        else "M"
        if not hic_algn2["is_unique"]
        else "U"
    )

    # Change orientation and positioning of pair for reporting:
    # AVAILABLE_REPORT_POSITION    = ["outer", "pair", "read", "walk"]
    # AVAILABLE_REPORT_ORIENTATION = ["pair", "pair", "read", "walk"]
    pair_type = pair_index[1]

    if report_orientation == "read":
        pass
    elif report_orientation == "walk":
        if pair_type == "R2":
            hic_algn1 = flip_orientation(hic_algn1)
            hic_algn2 = flip_orientation(hic_algn2)
        elif pair_type == "R1-2":
            hic_algn2 = flip_orientation(hic_algn2)
    elif report_orientation == "pair":
        if pair_type == "R1" or pair_type == "R1&R2":
            hic_algn2 = flip_orientation(hic_algn2)
        elif pair_type == "R2":
            hic_algn1 = flip_orientation(hic_algn1)
    elif report_orientation == "junction":
        if pair_type == "R1" or pair_type == "R1&R2":
            hic_algn1 = flip_orientation(hic_algn1)
        elif pair_type == "R2":
            hic_algn2 = flip_orientation(hic_algn2)
        else:
            hic_algn1 = flip_orientation(hic_algn1)
            hic_algn2 = flip_orientation(hic_algn2)

    if report_position == "read":
        pass
    elif report_position == "walk":
        if pair_type == "R2":
            hic_algn1 = flip_position(hic_algn1)
            hic_algn2 = flip_position(hic_algn2)
        elif pair_type == "R1-2":
            hic_algn2 = flip_position(hic_algn2)
    elif report_position == "outer":
        if pair_type == "R1" or pair_type == "R1&R2":
            hic_algn2 = flip_position(hic_algn2)
        elif pair_type == "R2":
            hic_algn1 = flip_position(hic_algn1)
    elif report_position == "junction":
        if pair_type == "R1" or pair_type == "R1&R2":
            hic_algn1 = flip_position(hic_algn1)
        elif pair_type == "R2":
            hic_algn2 = flip_position(hic_algn2)
        else:
            hic_algn1 = flip_position(hic_algn1)
            hic_algn2 = flip_position(hic_algn2)

    return [hic_algn1, hic_algn2, pair_index]


def check_pair_order(algn1, algn2, chrom_enum):
    """
    Check if a pair of alignments has the upper-triangular order or
    has to be flipped.
    """

    # First, the pair is flipped according to the type of mapping on its sides.
    # Later, we will check it is mapped on both sides and, if so, flip the sides
    # according to these coordinates.
    has_correct_order = (algn1["is_mapped"], algn1["is_unique"]) <= (
        algn2["is_mapped"],
        algn2["is_unique"],
    )

    # If a pair has coordinates on both sides, it must be flipped according to
    # its genomic coordinates.
    if (algn1["chrom"] != pairsam_format.UNMAPPED_CHROM) and (
        algn2["chrom"] != pairsam_format.UNMAPPED_CHROM
    ):
        has_correct_order = (chrom_enum[algn1["chrom"]], algn1["pos"]) <= (
            chrom_enum[algn2["chrom"]],
            algn2["pos"],
        )

    return has_correct_order


######################
### Output utilities:
######################


def write_all_algnments(readID, all_algns1, all_algns2, out_file):
    """
    Debug utility that outputs all alignments in .bam file before parsing walks/pairs
    """
    for side_idx, all_algns in enumerate((all_algns1, all_algns2)):
        out_file.write(readID)
        out_file.write("\t")
        out_file.write(str(side_idx + 1))
        out_file.write("\t")
        for algn in sorted(all_algns, key=lambda x: x["dist_to_5"]):
            out_file.write(algn["chrom"])
            out_file.write("\t")
            out_file.write(str(algn["pos5"]))
            out_file.write("\t")
            out_file.write(algn["strand"])
            out_file.write("\t")
            out_file.write(str(algn["mapq"]))
            out_file.write("\t")
            out_file.write(str(algn["cigar"]))
            out_file.write("\t")
            out_file.write(str(algn["dist_to_5"]))
            out_file.write("\t")
            out_file.write(str(algn["dist_to_5"] + algn["algn_read_span"]))
            out_file.write("\t")
            out_file.write(str(algn["matched_bp"]))
            out_file.write("\t")

        out_file.write("\n")


def write_pairsam(
    algn1,
    algn2,
    readID,
    pair_index,
    sams1,
    sams2,
    out_file,
    drop_readid,
    drop_seq,
    drop_sam,
    add_pair_index,
    add_columns,
):
    """
    Write output pairsam.
    Note: SAM is already tab-separated and
    any printable character between ! and ~ may appear in the PHRED field!
    (http://www.ascii-code.com/)
    Thus, use the vertical tab character to separate fields!
    """
    cols = [
        "." if drop_readid else readID,
        algn1["chrom"],
        str(algn1["pos"]),
        algn2["chrom"],
        str(algn2["pos"]),
        algn1["strand"],
        algn2["strand"],
        algn1["type"] + algn2["type"],
    ]

    if not drop_sam:
        for sams in [sams1, sams2]:
            if drop_seq:
                for sam in sams:
                    sam.query_qualities = ""
                    sam.query_sequence = ""
            cols.append(
                pairsam_format.INTER_SAM_SEP.join(
                    [
                        sam.to_string().replace(
                            "\t", pairsam_format.SAM_SEP
                        )  # String representation of pysam alignment
                        + pairsam_format.SAM_SEP
                        + "Yt:Z:"
                        + algn1["type"]
                        + algn2["type"]
                        for sam in sams
                    ]
                )
            )

    if add_pair_index:
        cols.append(str(pair_index[0]))
        cols.append(pair_index[1])

    for col in add_columns:
        # use get b/c empty alignments would not have sam tags (NM, AS, etc)
        cols.append(str(algn1.get(col, "")))
        cols.append(str(algn2.get(col, "")))

    out_file.write(pairsam_format.PAIRSAM_SEP.join(cols) + "\n")



#!/usr/bin/env python
# -*- coding: utf-8  -*-

import sys
import ast
import pathlib

from .._logging import get_logger

logger = get_logger()

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

import click

from ..lib.dedup import streaming_dedup, streaming_dedup_cython
from ..lib.stats import PairCounter

UTIL_NAME = "pairtools_dedup"


@cli.command()
@click.argument("pairs_path", type=str, required=False)

### Output files:
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file for pairs after duplicate removal."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@click.option(
    "--output-dups",
    type=str,
    default="",
    help="output file for duplicated pairs. "
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " If the path is the same as in --output or -, output duplicates together "
    " with deduped pairs. By default, duplicates are dropped.",
)
@click.option(
    "--output-unmapped",
    type=str,
    default="",
    help="output file for unmapped pairs. "
    "If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed. "
    "If the path is the same as in --output or -, output unmapped pairs together "
    "with deduped pairs. If the path is the same as --output-dups, output "
    "unmapped reads together with dups. By default, unmapped pairs are dropped.",
)
@click.option(
    "--output-stats",
    type=str,
    default="",
    help="output file for duplicate statistics."
    " If file exists, it will be open in the append mode."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, statistics are not printed.",
)
@click.option(
    "--output-bytile-stats",
    type=str,
    default="",
    help="output file for duplicate statistics."
    " Note that the readID should be provided and contain tile information for this option. "
    " This analysis is possible when pairtools is run on a dataset with original Illumina-generated read IDs, "
    " because SRA does not store original read IDs from the sequencer. "
    " By default, by-tile duplicate statistics are not printed. "
    " If file exists, it will be open in the append mode. "
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed.",
)

### Set the dedup method:
@click.option(
    "--max-mismatch",
    type=int,
    default=3,
    show_default=True,
    help="Pairs with both sides mapped within this distance (bp) from each "
    "other are considered duplicates. [dedup option]",
)
@click.option(
    "--method",
    type=click.Choice(["max", "sum"]),
    default="max",
    help="define the mismatch as either the max or the sum of the mismatches of"
    "the genomic locations of the both sides of the two compared molecules. [dedup option]",
    show_default=True,
)
@click.option(
    "--backend",
    type=click.Choice(["scipy", "sklearn", "cython"]),
    default="scipy",
    help="What backend to use: scipy and sklearn are based on KD-trees,"
    " cython is online indexed list-based algorithm."
    " With cython backend, duplication is not transitive with non-zero max mismatch "
    " (e.g. pairs A and B are duplicates, and B and C are duplicates, then A and C are "
    " not necessary duplicates of each other), while with scipy and sklearn it's "
    " transitive (i.e. A and C are necessarily duplicates)."
    " Cython is the original version used in pairtools since its beginning."
    " It is available for backwards compatibility and to allow specification of the"
    " column order."
    " Now the default scipy backend is generally the fastest, and with chunksize below"
    " 1 mln has the lowest memory requirements. [dedup option]"
    # " 'cython' is deprecated and provided for backwards compatibility",
)

### Scipy and sklearn-specific options:
@click.option(
    "--chunksize",
    type=int,
    default=100_000,
    show_default=True,
    help="Number of pairs in each chunk. Reduce for lower memory footprint."
    " Below 10,000 performance starts suffering significantly and the algorithm might"
    " miss a few duplicates with non-zero --max-mismatch."
    " Only works with '--backend scipy or sklearn'. [dedup option]",
)
@click.option(
    "--carryover",
    type=int,
    default=100,
    show_default=True,
    help="Number of deduped pairs to carry over from previous chunk to the new chunk"
    " to avoid breaking duplicate clusters."
    " Only works with '--backend scipy or sklearn'. [dedup option]",
)
@click.option(
    "-p",
    "--n-proc",
    type=int,
    default=1,
    help="Number of cores to use. Only applies with sklearn backend."
    "Still needs testing whether it is ever useful. [dedup option]",
)

### Output options:
@click.option(
    "--mark-dups",
    is_flag=True,
    help='If specified, duplicate pairs are marked as DD in "pair_type" and '
    "as a duplicate in the sam entries. [output format option]",
)
@click.option(
    "--keep-parent-id",
    is_flag=True,
    help="If specified, duplicate pairs are marked with the readID of the retained"
    " deduped read in the 'parent_readID' field. [output format option]",
)
@click.option(
    "--extra-col-pair",
    nargs=2,
    # type=click.Tuple([str, str]),
    multiple=True,
    help="Extra columns that also must match for two pairs to be marked as "
    "duplicates. Can be either provided as 0-based column indices or as column "
    'names (requires the "#columns" header field). The option can be provided '
    "multiple times if multiple column pairs must match. "
    'Example: --extra-col-pair "phase1" "phase2". [output format option]',
)

### Input options:
@click.option(
    "--sep",
    type=str,
    default=pairsam_format.PAIRSAM_SEP_ESCAPE,
    help=r"Separator (\t, \v, etc. characters are "
    "supported, pass them in quotes). [input format option]",
)
@click.option(
    "--send-header-to",
    type=click.Choice(["dups", "dedup", "both", "none"]),
    default="both",
    help="Which of the outputs should receive header and comment lines. [input format option]",
)
@click.option(
    "--c1",
    type=int,
    default=pairsam_format.COL_C1,
    help=f"Chrom 1 column; default {pairsam_format.COL_C1}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--c2",
    type=int,
    default=pairsam_format.COL_C2,
    help=f"Chrom 2 column; default {pairsam_format.COL_C2}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--p1",
    type=int,
    default=pairsam_format.COL_P1,
    help=f"Position 1 column; default {pairsam_format.COL_P1}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--p2",
    type=int,
    default=pairsam_format.COL_P2,
    help=f"Position 2 column; default {pairsam_format.COL_P2}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--s1",
    type=int,
    default=pairsam_format.COL_S1,
    help=f"Strand 1 column; default {pairsam_format.COL_S1}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--s2",
    type=int,
    default=pairsam_format.COL_S2,
    help=f"Strand 2 column; default {pairsam_format.COL_S2}"
    " Only works with '--backend cython'. [input format option]",
)
@click.option(
    "--unmapped-chrom",
    type=str,
    default=pairsam_format.UNMAPPED_CHROM,
    help="Placeholder for a chromosome on an unmapped side; default {}".format(
        pairsam_format.UNMAPPED_CHROM
    ),
)

# Output stats option
@click.option(
    "--yaml/--no-yaml",
    is_flag=True,
    default=False,
    help="Output stats in yaml format instead of table. [output stats format option]",
)

# Filtering options for reporting stats:
@click.option(
    "--filter",
    default=None,
    required=False,
    multiple=True,
    help="Filter stats with condition to apply to the data (similar to `pairtools select` or `pairtools stats`). "
    "For non-YAML output only the first filter will be reported. [output stats filtering option] "
    "Note that this will not change the deduplicated output pairs. "
    """Example: pairtools dedup --yaml --filter 'unique:(pair_type=="UU")' --filter 'close:(pair_type=="UU") and (abs(pos1-pos2)<10)' --output-stats - test.pairs """,
)
@click.option(
    "--engine",
    default="pandas",
    required=False,
    help="Engine for regular expression parsing for stats filtering. "
    "Python will provide you regex functionality, while pandas does not accept "
    "custom funtctions and works faster. [output stats filtering option]",
)
@click.option(
    "--chrom-subset",
    type=str,
    default=None,
    required=False,
    help="A path to a chromosomes file (tab-separated, 1st column contains "
    "chromosome names) containing a chromosome subset of interest for stats filter. "
    "If provided, additionally filter pairs with both sides originating from "
    "the provided subset of chromosomes. This operation modifies the #chromosomes: "
    "and #chromsize: header fields accordingly.  "
    "Note that this will not change the deduplicated output pairs. [output stats filtering option]",
)
@click.option(
    "--startup-code",
    type=str,
    default=None,
    required=False,
    help="An auxiliary code to execute before filteringfor stats. "
    "Use to define functions that can be evaluated in the CONDITION statement. [output stats filtering option]",
)
@click.option(
    "-t",
    "--type-cast",
    type=(str, str),
    default=(),
    multiple=True,
    help="Cast a given column to a given type for stats filtering. By default, only pos and mapq "
    "are cast to int, other columns are kept as str. Provide as "
    "-t <column_name> <type>, e.g. -t read_len1 int. Multiple entries are allowed. [output stats filtering option]",
)
@common_io_options
def dedup(
    pairs_path,
    output,
    output_dups,
    output_unmapped,
    output_stats,
    output_bytile_stats,
    chunksize,
    carryover,
    max_mismatch,
    method,
    sep,
    send_header_to,
    c1,
    c2,
    p1,
    p2,
    s1,
    s2,
    unmapped_chrom,
    mark_dups,
    extra_col_pair,
    keep_parent_id,
    backend,
    n_proc,
    **kwargs,
):
    """Find and remove PCR/optical duplicates.

    Find PCR/optical duplicates in an upper-triangular flipped sorted pairs/pairsam
    file. Allow for a +/-N bp mismatch at each side of duplicated molecules.

    PAIRS_PATH : input triu-flipped sorted .pairs or .pairsam file.  If the
    path ends with .gz/.lz4, the input is decompressed by bgzip/lz4c.
    By default, the input is read from stdin.
    """

    dedup_py(
        pairs_path,
        output,
        output_dups,
        output_unmapped,
        output_stats,
        output_bytile_stats,
        chunksize,
        carryover,
        max_mismatch,
        method,
        sep,
        send_header_to,
        c1,
        c2,
        p1,
        p2,
        s1,
        s2,
        unmapped_chrom,
        mark_dups,
        extra_col_pair,
        keep_parent_id,
        backend,
        n_proc,
        **kwargs,
    )


if __name__ == "__main__":
    dedup()


def dedup_py(
    pairs_path,
    output,
    output_dups,
    output_unmapped,
    output_stats,
    output_bytile_stats,
    chunksize,
    carryover,
    max_mismatch,
    method,
    sep,
    send_header_to,
    c1,
    c2,
    p1,
    p2,
    s1,
    s2,
    unmapped_chrom,
    mark_dups,
    extra_col_pair,
    keep_parent_id,
    backend,
    n_proc,
    **kwargs,
):

    sep = ast.literal_eval('"""' + sep + '"""')
    send_header_to_dedup = send_header_to in ["both", "dedup"]
    send_header_to_dup = send_header_to in ["both", "dups"]

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )
    out_stats_stream = fileio.auto_open(
        output_stats,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    bytile_dups = False
    if output_bytile_stats:
        out_bytile_stats_stream = fileio.auto_open(
            output_bytile_stats,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        bytile_dups = True
        if not keep_parent_id:
            logger.warning(
                "Force output --parent-readID because --output-bytile-stats provided."
            )
            keep_parent_id = True

    # generate empty PairCounter if stats output is requested:
    if output_stats:
        filter = kwargs.get("filter", None)
        # Define filters and their properties
        first_filter_name = "no_filter"  # default filter name for full output
        if filter is not None and len(filter) > 0:
            first_filter_name = filter[0].split(":", 1)[0]
            if len(filter) > 1 and not kwargs.get("yaml", False):
                logger.warn(
                    f"Output the first filter only in non-YAML output: {first_filter_name}"
                )

            filter = dict([f.split(":", 1) for f in filter])
        else:
            filter = None

        out_stat = PairCounter(
            bytile_dups=bytile_dups,
            filters=filter,
            startup_code=kwargs.get("startup_code", ""),  # for evaluation of filters
            type_cast=kwargs.get("type_cast", ()),  # for evaluation of filters
            engine=kwargs.get("engine", "pandas"),
        )
    else:
        out_stat = None

    if not output_dups:
        outstream_dups = None
    elif output_dups == "-" or (
        pathlib.Path(output_dups).absolute() == pathlib.Path(output).absolute()
    ):
        outstream_dups = outstream
    else:
        outstream_dups = fileio.auto_open(
            output_dups,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    if not output_unmapped:
        outstream_unmapped = None
    elif output_unmapped == "-" or (
        pathlib.Path(output_unmapped).absolute() == pathlib.Path(output).absolute()
    ):
        outstream_unmapped = outstream
    elif (
        pathlib.Path(output_unmapped).absolute() == pathlib.Path(output_dups).absolute()
    ):
        outstream_unmapped = outstream_dups
    else:
        outstream_unmapped = fileio.auto_open(
            output_unmapped,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    header, body_stream = headerops.get_header(instream)
    if not any([l.startswith("#sorted") for l in header]):
        logger.warning(
            "Pairs file appears not to be sorted, dedup might produce wrong results."
        )
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    if send_header_to_dedup:
        outstream.writelines((l + "\n" for l in header))
    if send_header_to_dup and outstream_dups and (outstream_dups != outstream):
        dups_header = header
        if keep_parent_id and len(dups_header) > 0:
            dups_header = headerops.append_columns(dups_header, ["parent_readID"])
        outstream_dups.writelines((l + "\n" for l in dups_header))
    if (
        outstream_unmapped
        and (outstream_unmapped != outstream)
        and (outstream_unmapped != outstream_dups)
    ):
        outstream_unmapped.writelines((l + "\n" for l in header))

    column_names = headerops.extract_column_names(header)
    extra_cols1 = []
    extra_cols2 = []
    if extra_col_pair is not None:
        for col1, col2 in extra_col_pair:
            extra_cols1.append(column_names[col1] if col1.isdigit() else col1)
            extra_cols2.append(column_names[col2] if col2.isdigit() else col2)

    if backend == "cython":
        # warnings.warn(
        #     "'cython' backend is deprecated and provided only"
        #     " for backwards compatibility",
        #     DeprecationWarning,
        # )
        extra_cols1 = [column_names.index(col) for col in extra_cols1]
        extra_cols2 = [column_names.index(col) for col in extra_cols2]
        streaming_dedup_cython(
            method,
            max_mismatch,
            sep,
            c1,
            c2,
            p1,
            p2,
            s1,
            s2,
            extra_cols1,
            extra_cols2,
            unmapped_chrom,
            body_stream,
            outstream,
            outstream_dups,
            outstream_unmapped,
            out_stat,
            mark_dups,
            keep_parent_id,
        )
    elif backend in ("scipy", "sklearn"):
        streaming_dedup(
            in_stream=body_stream,
            colnames=column_names,
            chunksize=chunksize,
            carryover=carryover,
            method=method,
            mark_dups=mark_dups,
            max_mismatch=max_mismatch,
            extra_col_pairs=list(extra_col_pair),
            keep_parent_id=keep_parent_id,
            unmapped_chrom=unmapped_chrom,
            outstream=outstream,
            outstream_dups=outstream_dups,
            outstream_unmapped=outstream_unmapped,
            out_stat=out_stat,
            backend=backend,
            n_proc=n_proc,
        )
    else:
        raise ValueError("Unknown backend")

    # save statistics to a file if it was requested:
    if out_stat:
        out_stat.save(
            out_stats_stream,
            yaml=kwargs.get("yaml", False),  # format as yaml
            filter=first_filter_name
            if not kwargs.get("yaml", False)
            else None,  # output only the first filter if non-YAML output
        )

    if bytile_dups:
        out_stat.save_bytile_dups(out_bytile_stats_stream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()

    if outstream_dups and (outstream_dups != outstream):
        outstream_dups.close()

    if (
        outstream_unmapped
        and (outstream_unmapped != outstream)
        and (outstream_unmapped != outstream_dups)
    ):
        outstream_unmapped.close()

    if out_stats_stream:
        out_stats_stream.close()



import numpy as np
import pandas as pd

import scipy.spatial
from scipy.sparse import coo_matrix
from scipy.sparse.csgraph import connected_components

from . import dedup_cython, pairsam_format
from .stats import PairCounter

from .._logging import get_logger

logger = get_logger()
import time

# Ignore pandas future warnings:
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# Setting for cython deduplication:
# you don't need to load more than 10k lines at a time b/c you get out of the
# CPU cache, so this parameter is not adjustable
MAX_LEN = 10000


def streaming_dedup(
    in_stream,
    colnames,
    chunksize,
    carryover,
    method,
    mark_dups,
    max_mismatch,
    extra_col_pairs,
    unmapped_chrom,
    outstream,
    outstream_dups,
    outstream_unmapped,
    keep_parent_id,
    out_stat,
    backend,
    n_proc,
):

    deduped_chunks = _dedup_stream(
        in_stream=in_stream,
        colnames=colnames,
        method=method,
        chunksize=chunksize,
        carryover=carryover,
        mark_dups=mark_dups,
        max_mismatch=max_mismatch,
        extra_col_pairs=extra_col_pairs,
        keep_parent_id=keep_parent_id,
        backend=backend,
        n_proc=n_proc,
    )

    t0 = time.time()
    N = 0

    for df_chunk in deduped_chunks:
        N += df_chunk.shape[0]

        # Write the stats if requested:
        if out_stat is not None:
            out_stat.add_pairs_from_dataframe(df_chunk, unmapped_chrom=unmapped_chrom)

        # Define masks of unmapped and duplicated reads:
        mask_mapped = np.logical_and(
            (df_chunk["chrom1"] != unmapped_chrom),
            (df_chunk["chrom2"] != unmapped_chrom),
        )
        mask_duplicates = df_chunk["duplicate"]

        # Clean up dataframe:
        df_chunk = df_chunk.drop(columns=["duplicate"])

        # Stream the dups:
        if outstream_dups:
            df_chunk.loc[mask_mapped & mask_duplicates, :].to_csv(
                outstream_dups, index=False, header=False, sep="\t"
            )

        # Drop readID if it was created (not needed for nodup and unmapped pairs):
        if keep_parent_id:
            df_chunk = df_chunk.drop(columns=["parent_readID"])

        # Stream unmapped:
        if outstream_unmapped:
            df_chunk.loc[~mask_mapped, :].to_csv(
                outstream_unmapped, index=False, header=False, sep="\t"
            )

        # Stream unique pairs:
        df_chunk.loc[mask_mapped & (~mask_duplicates), :].to_csv(
            outstream, index=False, header=False, sep="\t"
        )

    t1 = time.time()
    t = t1 - t0
    logger.debug(f"total time: {t}")
    if N > 0:
        logger.debug(f"time per mln pairs: {t/N*1e6}")
    else:
        logger.debug(f"Processed {N} pairs")


def _dedup_stream(
    in_stream,
    colnames,
    method,
    chunksize,
    carryover,
    mark_dups,
    max_mismatch,
    extra_col_pairs,
    keep_parent_id,
    backend,
    n_proc,
):
    # Stream the input dataframe:
    dfs = pd.read_table(
        in_stream, comment=None, names=colnames, chunksize=chunksize
    )

    # Set up the carryover dataframe:
    df_prev_nodups = pd.DataFrame([])
    prev_i = 0

    # Iterate over chunks:
    for df in dfs:
        df_marked = _dedup_chunk(
            pd.concat([df_prev_nodups, df], axis=0, ignore_index=True).reset_index(
                drop=True
            ),
            r=max_mismatch,
            method=method,
            keep_parent_id=keep_parent_id,
            extra_col_pairs=extra_col_pairs,
            backend=backend,
            n_proc=n_proc,
        )
        df_marked = df_marked.loc[prev_i:, :].reset_index(drop=True)
        mask_duplicated = df_marked["duplicate"]
        if mark_dups:
            df_marked.loc[mask_duplicated, "pair_type"] = "DD"

        # Filter out duplicates and store specific columns:
        df_nodups = df_marked.loc[~mask_duplicated, colnames]

        # Re-define carryover pairs:
        df_prev_nodups = df_nodups.tail(carryover).reset_index(drop=True)
        prev_i = len(df_prev_nodups)

        yield df_marked


def _dedup_chunk(
    df,
    r,
    method,
    keep_parent_id,
    extra_col_pairs,
    backend,
    unmapped_chrom="!",
    n_proc=1,
):
    """Mark duplicates in a dataframe of pairs

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with pairs, has to contain columns 'chrom1', 'pos1', 'chrom2', 'pos2'
        'strand1', 'strand2'
    r : int
        Allowed distance between two pairs to call them duplicates
    method : str
        'sum' or 'max' - whether 'r' uses sum of distances on two ends of pairs, or the
        maximal distance
    keep_parent_id : bool
        If True, the read ID of the read that was not labelled as a duplicate from a
        group of duplicates is recorded for each read marked as duplicate.
        Only possible with non-cython backends
    extra_col_pairs : list of tuples
        List of extra column pairs that need to match between two reads for them be
        considered duplicates (e.g. useful if alleles are annotated)
    backend : str
        'scipy', 'sklearn', 'cython'
    unmapped_chrom : str, optional
        Which character denotes unmapped reads in the chrom1/chrom2 fields,
        by default "!"
    n_proc : int, optional
        How many cores to use, by default 1
        Only works for 'sklearn' backend

    Returns
    -------
    pd.DataFrame
        Dataframe with marked duplicates (extra boolean field 'duplicate'), and
        optionally recorded 'parent_readID'

    """
    if method not in ("max", "sum"):
        raise ValueError('Unknown method, only "sum" or "max" allowed')
    if backend == "sklearn":
        from sklearn import neighbors

    if method == "sum":
        p = 1
    else:
        p = np.inf

    # Store the index of the dataframe:
    index_colname = df.index.name
    if index_colname is None:
        index_colname = "index"
    df = df.reset_index()  # Remove the index temporarily

    # Set up columns to store the dedup info:
    df.loc[:, "clusterid"] = np.nan
    df.loc[:, "duplicate"] = False

    # Split mapped and unmapped reads:
    mask_unmapped = (df["chrom1"] == unmapped_chrom) | (df["chrom2"] == unmapped_chrom)
    df_unmapped = df.loc[mask_unmapped, :].copy()
    df_mapped = df.loc[~mask_unmapped, :].copy()
    N_mapped = df_mapped.shape[0]

    # If there are some mapped reads, dedup them:
    if N_mapped > 0:
        if backend == "sklearn":
            a = neighbors.radius_neighbors_graph(
                df_mapped[["pos1", "pos2"]],
                radius=r,
                p=p,
                n_jobs=n_proc,
            )
            a0, a1 = a.nonzero()
        elif backend == "scipy":
            z = scipy.spatial.cKDTree(df_mapped[["pos1", "pos2"]])
            a = z.query_pairs(r=r, p=p, output_type="ndarray")
            a0 = a[:, 0]
            a1 = a[:, 1]
        need_to_match = np.array(
            [
                ("chrom1", "chrom1"),
                ("chrom2", "chrom2"),
                ("strand1", "strand1"),
                ("strand2", "strand2"),
            ]
            + extra_col_pairs
        )
        nonpos_matches = np.all(
            [
                df_mapped.iloc[a0, df_mapped.columns.get_loc(lc)].values
                == df_mapped.iloc[a1, df_mapped.columns.get_loc(rc)].values
                for (lc, rc) in need_to_match
            ],
            axis=0,
        )
        a0 = a0[nonpos_matches]
        a1 = a1[nonpos_matches]
        a_mat = coo_matrix((np.ones_like(a0), (a0, a1)), shape=(N_mapped, N_mapped))

        # Set up inferred clusterIDs:
        df_mapped.loc[:, "clusterid"] = connected_components(a_mat, directed=False)[1]

    mask_dups = df_mapped["clusterid"].duplicated()
    df_mapped.loc[mask_dups, "duplicate"] = True

    # Mark parent IDs if requested:
    if keep_parent_id:
        df_mapped.loc[:, "parent_readID"] = df_mapped["clusterid"].map(
            df_mapped[~mask_dups].set_index("clusterid")["readID"]
        )
        df_unmapped.loc[:, "parent_readID"] = ""

    # Reconstruct original dataframe with removed duplicated reads:
    df = pd.concat([df_unmapped, df_mapped]).reset_index(drop=True)
    df = df.set_index(index_colname)  # Set up the original index
    df = df.drop(
        ["clusterid"], axis=1
    )  # Remove the information that we don't need anymore:

    return df


### Cython deduplication ####
def streaming_dedup_cython(
    method,
    max_mismatch,
    sep,
    c1ind,
    c2ind,
    p1ind,
    p2ind,
    s1ind,
    s2ind,
    extra_cols1,
    extra_cols2,
    unmapped_chrom,
    instream,
    outstream,
    outstream_dups,
    outstream_unmapped,
    out_stat,
    mark_dups,
    keep_parent_id=False,
    readid_ind=0,
):
    """
    Cython-powered deduplication with online algorithm based on indexed list.

    Parameters
    ----------
    method: "max" or "sum"
    max_mismatch: maximum allowed mismatch to count the pairs as duplicates
    sep: separator of the fields in the input file
    c1ind: index of the chr1 column
    c2ind: index of the chr2 column
    p1ind: index of the pos1 column
    p2ind: index of the pos2 column
    s1ind: index of the strand1 column
    s2ind: index of the strand2 column
    extra_cols1: extra columns for left alignment in a pair to add
    extra_cols2: extra columns for right alignment in a pair to add
    unmapped_chrom: Symbol of the chromosome for the unmapped alignment
    instream: input stream of pair file
    outstream: output stram of deduplicated pairs
    outstream_dups: output stream of duplicates (optionally with added parent_id, see keep_parent_id option)
    outstream_unmapped: output stram of unmapped pairs
    out_stat: output statistics
    mark_dups: if True, will add "DD" as the pair_type
    keep_parent_id: if True, additional column "parent_readID will be added to the output, can be useful for optical duplicates search
    readid_ind: index of the readID column in the input file

    Returns
    -------

    """
    maxind = max(c1ind, c2ind, p1ind, p2ind, s1ind, s2ind)
    if bool(extra_cols1) and bool(extra_cols2):
        maxind = max(maxind, max(extra_cols1), max(extra_cols2))

    all_scols1 = [s1ind] + extra_cols1
    all_scols2 = [s2ind] + extra_cols2

    # if we do stats in the dedup, we need PAIR_TYPE
    # i do not see way around this:
    if out_stat:
        ptind = pairsam_format.COL_PTYPE
        maxind = max(maxind, ptind)

    dd = dedup_cython.OnlineDuplicateDetector(
        method, max_mismatch, returnData=False, keep_parent_id=keep_parent_id
    )

    c1 = []
    c2 = []
    p1 = []
    p2 = []
    s1 = []
    s2 = []
    idx = []
    line_buffer = []
    cols_buffer = []
    chromDict = {}
    strandDict = {}
    curMaxLen = max(MAX_LEN, dd.getLen())

    t0 = time.time()
    N = 0

    instream = iter(instream)
    read_idx = 0  # read index to mark the parent readID
    while True:
        rawline = next(instream, None)
        stripline = rawline.strip('\n') if rawline else None

        # take care of empty lines not at the end of the file separately
        if rawline and (not stripline):
            logger.warning("Empty line detected not at the end of the file")
            continue

        if stripline:
            cols = stripline.split(sep)
            if len(cols) <= maxind:
                raise ValueError(
                    "Error parsing line {}: ".format(stripline)
                    + " expected {} columns, got {}".format(maxind, len(cols))
                )

            if (cols[c1ind] == unmapped_chrom) or (cols[c2ind] == unmapped_chrom):

                if outstream_unmapped:
                    outstream_unmapped.write(stripline)
                    # don't forget terminal newline
                    outstream_unmapped.write("\n")

                # add a pair to PairCounter if stats output is requested:
                if out_stat:
                    out_stat.add_pair(
                        cols[c1ind],
                        int(cols[p1ind]),
                        cols[s1ind],
                        cols[c2ind],
                        int(cols[p2ind]),
                        cols[s2ind],
                        cols[ptind],
                    )
            else:
                line_buffer.append(stripline)
                cols_buffer.append(cols)

                c1.append(fetchadd(cols[c1ind], chromDict))
                c2.append(fetchadd(cols[c2ind], chromDict))
                p1.append(int(cols[p1ind]))
                p2.append(int(cols[p2ind]))

                idx.append(read_idx)
                read_idx += 1

                if bool(extra_cols1) and bool(extra_cols2):
                    s1.append(
                        fetchadd("".join(cols[i] for i in all_scols1), strandDict)
                    )
                    s2.append(
                        fetchadd("".join(cols[i] for i in all_scols2), strandDict)
                    )
                else:
                    s1.append(fetchadd(cols[s1ind], strandDict))
                    s2.append(fetchadd(cols[s2ind], strandDict))
            N += 1
        if (not stripline) or (len(c1) == curMaxLen):
            if keep_parent_id:
                res, parents = dd.push(
                    ar(c1, 32),
                    ar(c2, 32),
                    ar(p1, 32),
                    ar(p2, 32),
                    ar(s1, 32),
                    ar(s2, 32),
                )

            else:
                res = dd.push(
                    ar(c1, 32),
                    ar(c2, 32),
                    ar(p1, 32),
                    ar(p2, 32),
                    ar(s1, 32),
                    ar(s2, 32),
                )

            if not stripline:
                if keep_parent_id:
                    res_tmp, parents_tmp = dd.finish()
                    parents = np.concatenate([parents, parents_tmp])

                else:
                    res_tmp = dd.finish()
                res = np.concatenate([res, res_tmp])

            for i in range(len(res)):
                # not duplicated pair:
                if not res[i]:
                    outstream.write(line_buffer[i])
                    # don't forget terminal newline
                    outstream.write("\n")
                    if out_stat:
                        out_stat.add_pair(
                            cols_buffer[i][c1ind],
                            int(cols_buffer[i][p1ind]),
                            cols_buffer[i][s1ind],
                            cols_buffer[i][c2ind],
                            int(cols_buffer[i][p2ind]),
                            cols_buffer[i][s2ind],
                            cols_buffer[i][ptind],
                        )
                # duplicated pair:
                else:
                    if out_stat:
                        out_stat.add_pair(
                            cols_buffer[i][c1ind],
                            int(cols_buffer[i][p1ind]),
                            cols_buffer[i][s1ind],
                            cols_buffer[i][c2ind],
                            int(cols_buffer[i][p2ind]),
                            cols_buffer[i][s2ind],
                            "DD",
                        )
                    if outstream_dups:
                        if mark_dups:  # DD-marked pair:
                            output = sep.join(mark_split_pair_as_dup(cols_buffer[i]))
                        else:  # pair as is:
                            output = line_buffer[i]

                        if keep_parent_id:  # Add parentID as the last column:
                            parent_readID = line_buffer[parents[i]].split(sep)[
                                readid_ind
                            ]
                            output = sep.join([output, parent_readID])

                        outstream_dups.write(output)

                        # don't forget terminal newline
                        outstream_dups.write("\n")

            # flush buffers and perform necessary checks here:
            c1 = []
            c2 = []
            p1 = []
            p2 = []
            s1 = []
            s2 = []
            line_buffer = line_buffer[len(res) :]
            cols_buffer = cols_buffer[len(res) :]
            if not stripline:
                if len(line_buffer) != 0:
                    raise ValueError(
                        "{} lines left in the buffer, ".format(len(line_buffer))
                        + "should be none;"
                        + "something went terribly wrong"
                    )
                break
        # process next line ...

    # all lines have been processed at this point.
    # streaming_dedup is over.
    t1 = time.time()
    t = t1 - t0
    logger.debug(f"total time: {t}")
    if N > 0:
        logger.debug(f"time per mln pairs: {t/N*1e6}")
    else:
        logger.debug(f"Processed {N} pairs")


def fetchadd(key, mydict):
    key = key.strip()
    if key not in mydict:
        mydict[key] = len(mydict)
    return mydict[key]


def ar(mylist, val):
    return np.array(mylist, dtype={8: np.int8, 16: np.int16, 32: np.int32}[val])


#### Markasdup utilities: ####
def mark_split_pair_as_dup(cols):
    # if the original columns ended with a new line, the marked columns
    # should as well.
    original_has_newline = cols[-1].endswith("\n")

    cols[pairsam_format.COL_PTYPE] = "DD"

    if (len(cols) > pairsam_format.COL_SAM1) and (len(cols) > pairsam_format.COL_SAM2):
        for i in (pairsam_format.COL_SAM1, pairsam_format.COL_SAM2):

            # split each sam column into sam entries, tag and assemble back
            cols[i] = pairsam_format.INTER_SAM_SEP.join(
                [
                    mark_sam_as_dup(sam)
                    for sam in cols[i].split(pairsam_format.INTER_SAM_SEP)
                ]
            )

    if original_has_newline and not cols[-1].endswith("\n"):
        cols[-1] = cols[-1] + "\n"
    return cols


def mark_sam_as_dup(sam):
    """Tag the binary flag and the optional pair type field of a sam entry
    as a PCR duplicate."""
    samcols = sam.split(pairsam_format.SAM_SEP)

    if len(samcols) == 1:
        return sam

    samcols[1] = str(int(samcols[1]) | 1024)

    for j in range(11, len(samcols)):
        if samcols[j].startswith("Yt:Z:"):
            samcols[j] = "Yt:Z:DD"
    return pairsam_format.SAM_SEP.join(samcols)



from pysam.libcalignmentfile cimport AlignmentFile
from pysam.libcalignedsegment cimport AlignedSegment, AlignmentHeader
from pysam.libchtslib cimport *
from pysam.libcutils cimport array_to_qualitystring

cdef class AlignmentFilePairtoolized(AlignmentFile):
    """ Modified class that loads each entry as pairtoolozed alignment. """

    def __next__(self):
        cdef int ret = self.cnext()
        if (ret >= 0):
            # Redefine the constructed object:
            return makeAlignedSegmentPairtoolized(self.b, self.header)
        elif ret == -2:
            raise IOError('truncated file')
        else:
            raise StopIteration

cdef AlignedSegmentPairtoolized makeAlignedSegmentPairtoolized(bam1_t *src,
                                       AlignmentHeader header):
    '''return an AlignedSegmentPairtoolized object constructed from `src`'''
    # note that the following does not call __init__
    # Redefine the constructed object:
    cdef AlignedSegmentPairtoolized dest = AlignedSegmentPairtoolized.__new__(AlignedSegmentPairtoolized)
    dest._delegate = bam_dup1(src)
    dest.header = header
    return dest

cdef class AlignedSegmentPairtoolized(AlignedSegment):
    """ In the pairtoolized class we inherit everything and
        add some useful properties and functions on top of that.
    """

    def is_unique(self, min_mapq):
        """true if read is unique mapping (by mapq)"""
        return self.mapq >= min_mapq

    property is_linear:
        """true if read is linear (SA is present in tages)"""
        def __get__(self):

            if self.has_tag('SA'):
                return False
            # for tag in self.tags:
            #     if 'SA'==tag[0]:
            #         return False
            return True

    property cigar_dict:
        """Parsed CIGAR as dictionary with interpretable fields"""

        def __get__(self):
            """Parse cigar tuples reported as cigartuples of pysam read entry.
            Reports alignment span, clipped nucleotides and more.
            See https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment.cigartuples
            """
            matched_bp = 0
            algn_ref_span = 0
            algn_read_span = 0
            read_len = 0
            clip5_ref = 0
            clip3_ref = 0

            cigarstring = self.cigarstring
            cigartuples = self.cigartuples
            if cigartuples is not None:
                for operation, length in cigartuples:
                    if operation == 0:  # M, match
                        matched_bp += length
                        algn_ref_span += length
                        algn_read_span += length
                        read_len += length
                    elif operation == 1:  # I, insertion
                        algn_read_span += length
                        read_len += length
                    elif operation == 2:  # D, deletion
                        algn_ref_span += length
                    elif (
                            operation == 4 or operation == 5
                    ):  # S and H, soft clip and hard clip, respectively
                        read_len += length
                        if matched_bp == 0:
                            clip5_ref = length
                        else:
                            clip3_ref = length

            return {
                "clip5_ref": clip5_ref,
                "clip3_ref": clip3_ref,
                "cigar": cigarstring,
                "algn_ref_span": algn_ref_span,
                "algn_read_span": algn_read_span,
                "read_len": read_len,
                "matched_bp": matched_bp
            }


from cpython cimport array
import cython
cimport cython

cpdef list get_mismatches_c(str seq, array.array quals, list aligned_pairs):
    '''
    This function takes a SAM alignment and, for every mismatch between the read and reference sequences,
    returns a tuple (the reference bp, the read bp, PHRED quality of the bp, reference position, read position).
    
    Reference: https://github.com/gerlichlab/scshic_pipeline/blob/master/bin/seq_mismatches.pyx 
    '''

    cdef cython.int read_pos, ref_pos
    cdef str orig_bp, orig_bp_upper
    cdef list mismatches = []

    for read_pos, ref_pos, orig_bp in aligned_pairs:
        orig_bp_upper = orig_bp.upper()
        if (seq[read_pos] != orig_bp_upper):
            mismatches.append(
                (orig_bp_upper,
                 seq[read_pos],
                 quals[read_pos],
                 ref_pos,
                 read_pos)
            )

    return mismatches


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import click
import sys

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.stats import PairCounter
from ..lib.parse_pysam import AlignmentFilePairtoolized
from ..lib.parse import streaming_classify


UTIL_NAME = "pairtools_parse"


@cli.command()
@click.argument("sam_path", type=str, required=False)
@click.option(
    "-c",
    "--chroms-path",
    type=str,
    required=True,
    help="Chromosome order used to flip interchromosomal mates: "
    "path to a chromosomes file (e.g. UCSC chrom.sizes or similar) whose "
    "first column lists scaffold names. Any scaffolds not listed will be "
    "ordered lexicographically following the names provided.",
)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file. "
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4-compressed."
    "By default, the output is printed into stdout. ",
)
@click.option(
    "--assembly",
    type=str,
    help="Name of genome assembly (e.g. hg19, mm10) to store in the pairs header.",
)
@click.option(
    "--min-mapq",
    type=int,
    default=1,
    show_default=True,
    help="The minimal MAPQ score to consider a read as uniquely mapped",
)
@click.option(
    "--max-molecule-size",
    type=int,
    default=750,
    show_default=True,
    help="The maximal size of a Hi-C molecule; used to rescue single ligations"
    "(from molecules with three alignments) and to rescue complex ligations."
    "The default is based on oriented P(s) at short ranges of multiple Hi-C."
    "Not used with walks-policy all.",
)
@click.option(
    "--drop-readid",
    is_flag=True,
    help="If specified, do not add read ids to the output",
)
@click.option(
    "--drop-seq",
    is_flag=True,
    help="If specified, remove sequences and PHREDs from the sam fields",
)
@click.option(
    "--drop-sam", is_flag=True, help="If specified, do not add sams to the output"
)
@click.option(
    "--add-pair-index",
    is_flag=True,
    help="If specified, each pair will have pair index in the molecule",
)
@click.option(
    "--add-columns",
    type=click.STRING,
    default="",
    help="Report extra columns describing alignments "
    "Possible values (can take multiple values as a comma-separated "
    "list): a SAM tag (any pair of uppercase letters) or {}.".format(
        ", ".join(pairsam_format.EXTRA_COLUMNS)
    ),
)
@click.option(
    "--output-parsed-alignments",
    type=str,
    default="",
    help="output file for all parsed alignments, including walks."
    " Useful for debugging and rnalysis of walks."
    " If file exists, it will be open in the append mode."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4-compressed."
    " By default, not used.",
)
@click.option(
    "--output-stats",
    type=str,
    default="",
    help="output file for various statistics of pairs file. "
    " By default, statistics is not generated.",
)
@click.option(
    "--report-alignment-end",
    type=click.Choice(["5", "3"]),
    default="5",
    help="specifies whether the 5' or 3' end of the alignment is reported as"
    " the position of the Hi-C read.",
)
@click.option(
    "--max-inter-align-gap",
    type=int,
    default=20,
    show_default=True,
    help="read segments that are not covered by any alignment and"
    ' longer than the specified value are treated as "null" alignments.'
    " These null alignments convert otherwise linear alignments into walks,"
    " and affect how they get reported as a Hi-C pair (see --walks-policy).",
)
@click.option(
    "--walks-policy",
    type=click.Choice(["mask", "5any", "5unique", "3any", "3unique", "all"]),
    default="mask",
    help="the policy for reporting unrescuable walks (reads containing more"
    " than one alignment on one or both sides, that can not be explained by a"
    " single ligation between two mappable DNA fragments)."
    ' "mask" - mask walks (chrom="!", pos=0, strand="-"); '
    ' "5any" - report the 5\'-most alignment on each side;'
    ' "5unique" - report the 5\'-most unique alignment on each side, if present;'
    ' "3any" - report the 3\'-most alignment on each side;'
    ' "3unique" - report the 3\'-most unique alignment on each side, if present;'
    ' "all" - report all available unique alignments on each side.',
    show_default=True,
)
@click.option(
    "--readid-transform",
    type=str,
    default=None,
    help="A Python expression to modify read IDs. Useful when read IDs differ "
    "between the two reads of a pair. Must be a valid Python expression that "
    "uses variables called readID and/or i (the 0-based index of the read pair "
    "in the bam file) and returns a new value, e.g. \"readID[:-2]+'_'+str(i)\". "
    "Make sure that transformed readIDs remain unique!",
    show_default=True,
)
@click.option(
    "--flip/--no-flip",
    is_flag=True,
    default=True,
    help="If specified, do not flip pairs in genomic order and instead preserve "
    "the order in which they were sequenced.",
)
@common_io_options
def parse(
    sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
):
    """Find ligation pairs in .sam data, make .pairs.
    SAM_PATH : an input .sam/.bam file with paired-end sequence alignments of
    Hi-C molecules. If the path ends with .bam, the input is decompressed from
    bam with samtools. By default, the input is read from stdin.
    """
    parse_py(
        sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
    )


def parse_py(
    sam_path, chroms_path, output, output_parsed_alignments, output_stats, **kwargs
):

    ### Set up input stream
    if sam_path:  # open input sam file with pysam
        input_sam = AlignmentFilePairtoolized(
            sam_path, "r", threads=kwargs.get("nproc_in")
        )
    else:  # read from stdin
        input_sam = AlignmentFilePairtoolized("-", "r", threads=kwargs.get("nproc_in"))

    ### Set up output streams
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    out_alignments_stream, out_stats_stream = None, None
    if output_parsed_alignments:
        out_alignments_stream = fileio.auto_open(
            output_parsed_alignments,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
    if output_stats:
        out_stats_stream = fileio.auto_open(
            output_stats,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    if out_alignments_stream:
        out_alignments_stream.write(
            "readID\tside\tchrom\tpos\tstrand\tmapq\tcigar\tdist_5_lo\tdist_5_hi\tmatched_bp\n"
        )

    # generate empty PairCounter if stats output is requested:
    out_stat = PairCounter() if output_stats else None

    ### Set up output parameters
    add_columns = kwargs.get("add_columns", [])
    add_columns = [col for col in add_columns.split(",") if col]
    for col in add_columns:
        if not (
            (col in pairsam_format.EXTRA_COLUMNS) or (len(col) == 2 and col.isupper())
        ):
            raise Exception("{} is not a valid extra column".format(col))

    columns = pairsam_format.COLUMNS + (
        [c + side for c in add_columns for side in ["1", "2"]]
    )

    if kwargs.get("drop_sam", True):
        columns.pop(columns.index("sam1"))
        columns.pop(columns.index("sam2"))

    if not kwargs.get("add_pair_index", False):
        columns.pop(columns.index("walk_pair_index"))
        columns.pop(columns.index("walk_pair_type"))

    ### Parse header
    samheader = input_sam.header

    if not samheader:
        raise ValueError(
            "The input sam is missing a header! If reading a bam file, please use `samtools view -h` to include the header."
        )

    ### Parse chromosome files present in the input
    sam_chromsizes = headerops.get_chromsizes_from_pysam_header(samheader)
    chromosomes = headerops.get_chrom_order(chroms_path, list(sam_chromsizes.keys()))

    ### Write new header to the pairsam file
    header = headerops.make_standard_pairsheader(
        assembly=kwargs.get("assembly", ""),
        chromsizes=[(chrom, sam_chromsizes[chrom]) for chrom in chromosomes],
        columns=columns,
        shape="whole matrix" if not kwargs["flip"] else "upper triangle",
    )

    header = headerops.insert_samheader_pysam(header, samheader)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))

    ### Parse input and write to the outputs
    streaming_classify(
        input_sam, outstream, chromosomes, out_alignments_stream, out_stat, **kwargs
    )

    # save statistics to a file if it was requested:
    if out_stat:
        out_stat.save(out_stats_stream)

    if outstream != sys.stdout:
        outstream.close()
    # close optional output streams if needed:
    if out_alignments_stream and out_alignments_stream != sys.stdout:
        out_alignments_stream.close()
    if out_stats_stream and out_stats_stream != sys.stdout:
        out_stats_stream.close()


if __name__ == "__main__":
    parse()
